import pytest
from pytest import raises as assert_raises

import cupy

from cupy import testing
from cupyx.scipy import signal
from cupyx.scipy.signal import iirdesign

try:
    import scipy.signal  # NOQA
except ImportError:
    pass


nimpl = pytest.mark.xfail(reason="not implemented")
prec_loss = pytest.mark.xfail(reason="zpk2tf loses precision")


@testing.with_requires("scipy>=1.8")
class TestIIRFilter:

    # NB: test_symmetry with higher order ellip filters need low tolerance
    # on older CUDA versions.

    @pytest.mark.parametrize("N", list(range(1, 25)))
    @pytest.mark.parametrize("ftype", ['butter',
                                       pytest.param('bessel', marks=nimpl),
                                       'cheby1', 'cheby2', 'ellip'])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-5, rtol=1e-6)
    def test_symmetry(self, N, ftype, xp, scp):
        # All built-in IIR filters are real, so should have perfectly
        # symmetrical poles and zeros. Then ba representation (using
        # numpy.poly) will be purely real instead of having negligible
        # imaginary parts.
        z, p, k = scp.signal.iirfilter(N, 1.1, 1, 20, 'low', analog=True,
                                       ftype=ftype, output='zpk')
        return z, p, k

    @pytest.mark.parametrize("N", list(range(1, 25)))
    @pytest.mark.parametrize("ftype", ['butter',
                                       pytest.param('bessel', marks=nimpl),
                                       'cheby1', 'cheby2', 'ellip'])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-5, rtol=1e-5)
    def test_symmetry_2(self, N, ftype, xp, scp):
        b, a = scp.signal.iirfilter(N, 1.1, 1, 20, 'low', analog=True,
                                    ftype=ftype, output='ba')
        return b, a

    @pytest.mark.xfail(reason="bessel IIR filter not implemented")
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_int_inputs(self, xp, scp):
        # Using integer frequency arguments and large N should not produce
        # numpy integers that wraparound to negative numbers
        z, p, k = scp.signal.iirfilter(24, 100, btype='low', analog=True,
                                       ftype='bessel', output='zpk')
        return z, p, k

    def test_invalid_wn_size(self):
        # low and high have 1 Wn, band and stop have 2 Wn
        assert_raises(ValueError, signal.iirfilter, 1, [0.1, 0.9], btype='low')
        assert_raises(ValueError, signal.iirfilter,
                      1, [0.2, 0.5], btype='high')
        assert_raises(ValueError, signal.iirfilter, 1, 0.2, btype='bp')
        assert_raises(ValueError, signal.iirfilter,
                      1, 400, btype='bs', analog=True)

    def test_invalid_wn_range(self):
        # For digital filters, 0 <= Wn <= 1
        assert_raises(ValueError, signal.iirfilter, 1, 2, btype='low')
        assert_raises(ValueError, signal.iirfilter, 1, [0.5, 1], btype='band')
        assert_raises(ValueError, signal.iirfilter, 1, [0., 0.5], btype='band')
        assert_raises(ValueError, signal.iirfilter, 1, -1, btype='high')
        assert_raises(ValueError, signal.iirfilter, 1, [1, 2], btype='band')
        assert_raises(ValueError, signal.iirfilter, 1, [10, 20], btype='stop')

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_analog_sos(self, xp, scp):
        # first order Butterworth filter with Wn = 1 has tf 1/(s+1)
        sos2 = scp.signal.iirfilter(
            N=1, Wn=1, btype='low', analog=True, output='sos')
        return sos2

    def test_wn1_ge_wn0(self):
        # gh-15773: should raise error if Wn[0] >= Wn[1]
        with pytest.raises(ValueError,
                           match=r"Wn\[0\] must be less than Wn\[1\]"):
            signal.iirfilter(2, [0.5, 0.5])
        with pytest.raises(ValueError,
                           match=r"Wn\[0\] must be less than Wn\[1\]"):
            signal.iirfilter(2, [0.6, 0.5])


@testing.with_requires("scipy")
class TestButter:

    @pytest.mark.parametrize('arg', [(0, 1), (1, 1)])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_degenerate(self, xp, scp, arg):
        # 0-order filter is just a passthrough
        b, a = scp.signal.butter(*arg, analog=True)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_degenerate_1(self, xp, scp):
        z, p, k = scp.signal.butter(1, 0.3, output='zpk')
        return z, p, k

    @pytest.mark.parametrize('N', list(range(25)))
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp, N):
        # analog s-plane
        wn = 0.01
        z, p, k = scp.signal.butter(N, wn, 'low', analog=True, output='zpk')
        return z, p, k

    @pytest.mark.parametrize('N', list(range(25)))
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_1(self, xp, scp, N):
        # digital z-plane
        wn = 0.01
        z, p, k = scp.signal.butter(N, wn, 'high', analog=False, output='zpk')
        return z, p, k

    @pytest.mark.parametrize('arg, analog',
                             [((2, 1), True),
                              ((5, 1), True),
                                 ((10, 1), True),
                                 ((19, 1.0441379169150726), True),
                                 ((5, 0.4), False)
                              ])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_2(self, xp, scp, arg, analog):
        b, a = scp.signal.butter(*arg, analog=analog)
        return b, a

    @pytest.mark.parametrize('arg', [(28, 0.43), (27, 0.56)])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_highpass(self, xp, scp, arg):
        # highpass, high even order
        z, p, k = scp.signal.butter(*arg, 'high', output='zpk')
        return z, p, k

    @pytest.mark.parametrize("format", ['zpk', 'ba'])
    @testing.numpy_cupy_allclose(scipy_name="scp", atol=1e-12)
    def test_bandpass(self, xp, scp, format):
        output = scp.signal.butter(8, [0.25, 0.33], 'band', output=format)
        return output

    @testing.numpy_cupy_allclose(scipy_name="scp", atol=1e-12)
    def test_bandpass_analog(self, xp, scp):
        output = scp.signal.butter(
            4, [90.5, 110.5], 'bp', analog=True, output='zpk')
        return output

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_bandstop(self, xp, scp):
        z, p, k = scp.signal.butter(7, [0.45, 0.56], 'stop', output='zpk')
        z.sort()
        p.sort()
        return z, p, k

    @pytest.mark.parametrize('outp',
                             ['zpk',
                              'sos',
                              pytest.param('ba', marks=prec_loss)])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    @testing.with_requires("scipy>=1.8")
    def test_ba_output(self, xp, scp, outp):
        outp = scp.signal.butter(
            4, [100, 300], 'bandpass', analog=True, output=outp)
        return outp

        # in scipy.signal, this is the output in the ba format.
        # in CUDA, zpk2tf loses precision and b,a output is garbage
        #  b2 = [1.6e+09, 0, 0, 0, 0]
        #  a2 = [1.000000000000000e+00, 5.226251859505511e+02,
        #        2.565685424949238e+05, 6.794127417357160e+07,
        #        1.519411254969542e+10, 2.038238225207147e+12,
        #        2.309116882454312e+14, 1.411088002066486e+16,
        #        8.099999999999991e+17]

    def test_fs_param(self):
        for fs in (900, 900.1, 1234.567):
            for N in (0, 1, 2, 3, 10):
                for fc in (100, 100.1, 432.12345):
                    for btype in ('lp', 'hp'):
                        ba1 = signal.butter(N, fc, btype, fs=fs)
                        ba2 = signal.butter(N, fc/(fs/2), btype)
                        testing.assert_allclose(ba1[0], ba2[0])
                        testing.assert_allclose(ba1[1], ba2[1])

                for fc in ((100, 200), (100.1, 200.2), (321.123, 432.123)):
                    for btype in ('bp', 'bs'):
                        ba1 = signal.butter(N, fc, btype, fs=fs)
                        # for seq in (list, tuple, array):
                        fcnorm = cupy.array([f/(fs/2) for f in fc])
                        ba2 = signal.butter(N, fcnorm, btype)
                        testing.assert_allclose(ba1[0], ba2[0])
                        testing.assert_allclose(ba1[0], ba2[0])


@testing.with_requires("scipy")
class TestCheby1:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate(self, xp, scp):
        # 0-order filter is just a passthrough
        # Even-order filters have DC gain of -rp dB
        b, a = scp.signal.cheby1(0, 10*xp.log10(2), 1, analog=True)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate_1(self, xp, scp):
        # 1-order filter is same for all types
        b, a = scp.signal.cheby1(1, 10*xp.log10(2), 1, analog=True)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate_2(self, xp, scp):
        z, p, k = scp.signal.cheby1(1, 0.1, 0.3, output='zpk')
        return z, p, k

    @pytest.mark.parametrize("N", list(range(25)))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic(self, xp, scp, N):
        wn = 0.01
        z, p, k = scp.signal.cheby1(N, 1, wn, 'low', analog=True, output='zpk')
        return z, p, k

    @pytest.mark.parametrize("N", list(range(25)))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_1(self, xp, scp, N):
        wn = 0.01
        z, p, k = scp.signal.cheby1(
            N, 1, wn, 'high', analog=False, output='zpk')
        return z, p, k

    @pytest.mark.parametrize("arg, kwd",
                             [((8, 0.5, 0.048), {}),
                              ((4, 1, [0.4, 0.7]), {'btype': 'band'}),
                                 ((5, 3, 1), {'analog': True}),
                                 ((8, 0.5, 0.1), {}),
                                 ((8, 0.5, 0.25), {}),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_basic_2(self, xp, scp, arg, kwd):
        # Same test as TestNormalize
        b, a = scp.signal.cheby1(*arg, **kwd)
        return b, a

    @pytest.mark.parametrize("arg, kwd",
                             # high even order
                             [((24, 0.7, 0.2), {'output': 'zpk'}),
                              # high odd order
                              ((23, 0.8, 0.3), {'output': 'zpk'}),
                                 ((10, 1, 1000), {
                                     'analog': True, 'output': 'zpk'}),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_highpass(self, xp, scp, arg, kwd):
        # high even order
        z, p, k = scp.signal.cheby1(*arg, 'high', **kwd)
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandpass(self, xp, scp):
        z, p, k = scp.signal.cheby1(8, 1, [0.3, 0.4], 'bp', output='zpk')
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandstop(self, xp, scp):
        z, p, k = scp.signal.cheby1(7, 1, [0.5, 0.6], 'stop', output='zpk')
        z = z[xp.argsort(z.imag)]
        p = p[xp.argsort(p.imag)]
        return z, p, k

    @pytest.mark.xfail(reason='zpk2tf loses precision (cf TestButter)')
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-10)
    def test_ba_output(self, xp, scp,):
        # with transfer function conversion,  without digital conversion
        b, a = scp.signal.cheby1(5, 0.9, [210, 310], 'stop', analog=True)
        return b, a

#        b2 = [1.000000000000006e+00, 0,
#              3.255000000000020e+05, 0,
#              4.238010000000026e+10, 0,
#              2.758944510000017e+15, 0,
#              8.980364380050052e+19, 0,
#              1.169243442282517e+24
#              ]
#        a2 = [1.000000000000000e+00, 4.630555945694342e+02,
#              4.039266454794788e+05, 1.338060988610237e+08,
#              5.844333551294591e+10, 1.357346371637638e+13,
#              3.804661141892782e+15, 5.670715850340080e+17,
#              1.114411200988328e+20, 8.316815934908471e+21,
#              1.169243442282517e+24
#              ]
#        assert_allclose(b, b2, rtol=1e-14)
#        assert_allclose(a, a2, rtol=1e-14)


@testing.with_requires("scipy")
class TestCheby2:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate(self, xp, scp):
        # 0-order filter is just a passthrough
        # Stopband ripple factor doesn't matter
        b, a = scp.signal.cheby2(0, 123.456, 1, analog=True)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate_1(self, xp, scp):
        # 1-order filter is same for all types
        b, a = scp.signal.cheby2(1, 10*xp.log10(2), 1, analog=True)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate_2(self, xp, scp):
        z, p, k = scp.signal.cheby2(1, 50, 0.3, output='zpk')
        return z, p, k

    @pytest.mark.parametrize("N", list(range(25)))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic(self, xp, scp, N):
        wn = 0.01
        z, p, k = scp.signal.cheby2(
            N, 40, wn, 'low', analog=True, output='zpk')
        return z, p, k
        #    assert_(len(p) == N)
        #    assert_(all(np.real(p) <= 0))  # No poles in right half of S-plane

    @pytest.mark.parametrize("N", list(range(25)))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_1(self, xp, scp, N):
        wn = 0.01
        z, p, k = scp.signal.cheby2(
            N, 40, wn, 'high', analog=False, output='zpk')
        return z, p, k
        #    assert_(all(np.abs(p) <= 1))  # No poles outside unit circle

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_2(self, xp, scp):
        B, A = scp.signal.cheby2(18, 100, 0.5)
        return B, A

    @pytest.mark.parametrize("arg, kwd",
                             # high even order
                             [((26, 60, 0.3), {'output': 'zpk'}),
                              # high odd order
                              ((25, 80, 0.5), {'output': 'zpk'}),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_highpass(self, xp, scp, arg, kwd):
        # high even order
        z, p, k = scp.signal.cheby2(*arg, 'high', **kwd)
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandpass(self, xp, scp):
        z, p, k = scp.signal.cheby2(9, 40, [0.07, 0.2], 'pass', output='zpk')
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_bandstop(self, xp, scp):
        z, p, k = scp.signal.cheby2(6, 55, [0.1, 0.9], 'stop', output='zpk')
        z = z[xp.argsort(xp.angle(z))]
        p = p[xp.argsort(xp.angle(p))]
        return z, p, k

    @pytest.mark.xfail(reason='zpk2tf loses precision')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_ba_output(self, xp, scp):
        # with transfer function conversion, without digital conversion
        b, a = scp.signal.cheby2(5, 20, [2010, 2100], 'stop', True)
        return b, a
        # b2 = [1.000000000000000e+00, 0,  # Matlab: 6.683253076978249e-12,
        #       2.111512500000000e+07, 0,  # Matlab: 1.134325604589552e-04,
        #       1.782966433781250e+14, 0,  # Matlab: 7.216787944356781e+02,
        #       7.525901316990656e+20, 0,  # Matlab: 2.039829265789886e+09,
        #       1.587960565565748e+27, 0,  # Matlab: 2.161236218626134e+15,
        #       1.339913493808585e+33]
        # a2 = [1.000000000000000e+00, 1.849550755473371e+02,
        #       2.113222918998538e+07, 3.125114149732283e+09,
        #       1.785133457155609e+14, 1.979158697776348e+16,
        #       7.535048322653831e+20, 5.567966191263037e+22,
        #       1.589246884221346e+27, 5.871210648525566e+28,
        #       1.339913493808590e+33]
        # assert_allclose(b, b2, rtol=1e-14)
        # assert_allclose(a, a2, rtol=1e-14)


@testing.with_requires("scipy>=1.8")
class TestEllip:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate(self, xp, scp):
        # 0-order filter is just a passthrough
        # Even-order filters have DC gain of -rp dB
        # Stopband ripple factor doesn't matter
        b, a = scp.signal.ellip(0, 10*xp.log10(2), 123.456, 1, analog=True)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate_1(self, xp, scp):
        # 1-order filter is same for all types
        b, a = scp.signal.ellip(1, 10*xp.log10(2), 1, 1, analog=True)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_degenerate_2(self, xp, scp):
        z, p, k = scp.signal.ellip(1, 1, 55, 0.3, output='zpk')
        return z, p, k

    @pytest.mark.parametrize('N', list(range(25)))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic(self, xp, scp, N):
        wn = 0.01
        z, p, k = scp.signal.ellip(
            N, 1, 40, wn, 'low', analog=True, output='zpk')
        return z, p, k

    @pytest.mark.parametrize('N', list(range(20)))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_1(self, xp, scp, N):
        wn = 0.01
        z, p, k = scp.signal.ellip(
            N, 1, 40, wn, 'high', analog=False, output='zpk')
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14, rtol=1e-14)
    def test_basic_2(self, xp, scp):
        b3, a3 = scp.signal.ellip(5, 3, 26, 1, analog=True)
        return b3, a3

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_3(self, xp, scp):
        b, a = scp.signal.ellip(3, 1, 60, [0.4, 0.7], 'stop')
        return b, a

    @pytest.mark.parametrize("arg",
                             # high even order
                             [(24, 1, 80, 0.3, 'high'),
                              # high odd order
                              (23, 1, 70, 0.5, 'high'),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_highpass(self, xp, scp, arg):
        # high even order
        z, p, k = scp.signal.ellip(*arg, output='zpk')
        return z, p, k

    @pytest.mark.parametrize("arg",
                             [(7, 1, 40, [0.07, 0.2], 'pass'),
                              (5, 1, 75, [90.5, 110.5], 'pass', True),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandpass(self, xp, scp, arg):
        z, p, k = scp.signal.ellip(7, 1, 40, [0.07, 0.2], 'pass', output='zpk')
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandstop(self, xp, scp):
        z, p, k = scp.signal.ellip(8, 1, 65, [0.2, 0.4], 'stop', output='zpk')
        z = z[xp.argsort(xp.angle(z))]
        p = p[xp.argsort(xp.angle(p))]
        return z, p, k

    @pytest.mark.xfail(reason='zpk2tf loses precision')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_ba_output(self, xp, scp):
        # with transfer function conversion,  without digital conversion
        b, a = scp.signal.ellip(5, 1, 40, [201, 240], 'stop', True)
        return b, a

#        b2 = [
#             1.000000000000000e+00, 0,  # Matlab: 1.743506051190569e-13,
#             2.426561778314366e+05, 0,  # Matlab: 3.459426536825722e-08,
#             2.348218683400168e+10, 0,  # Matlab: 2.559179747299313e-03,
#             1.132780692872241e+15, 0,  # Matlab: 8.363229375535731e+01,
#             2.724038554089566e+19, 0,  # Matlab: 1.018700994113120e+06,
#             2.612380874940186e+23
#             ]
#        a2 = [
#             1.000000000000000e+00, 1.337266601804649e+02,
#             2.486725353510667e+05, 2.628059713728125e+07,
#             2.436169536928770e+10, 1.913554568577315e+12,
#             1.175208184614438e+15, 6.115751452473410e+16,
#             2.791577695211466e+19, 7.241811142725384e+20,
#             2.612380874940182e+23
#             ]
#        assert_allclose(b, b2, rtol=1e-6)
#        assert_allclose(a, a2, rtol=1e-4)


@testing.with_requires("scipy")
class TestButtord:

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-8)
    def test_lowpass(self, xp, scp):
        wp = 0.2
        ws = 0.3
        rp = 3
        rs = 60
        N, Wn = scp.signal.buttord(wp, ws, rp, rs, False)
        b, a = scp.signal.butter(N, Wn, 'lowpass', False)
        w, h = scp.signal.freqz(b, a)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-10)
    def test_highpass(self, xp, scp):
        wp = 0.3
        ws = 0.2
        rp = 3
        rs = 70
        N, Wn = scp.signal.buttord(wp, ws, rp, rs, False)
        b, a = scp.signal.butter(N, Wn, 'highpass', False)
        w, h = scp.signal.freqz(b, a)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandpass(self, xp, scp):
        wp = [0.2, 0.5]
        ws = [0.1, 0.6]
        rp = 3
        rs = 80
        N, Wn = scp.signal.buttord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandstop(self, xp, scp):
        wp = [0.1, 0.6]
        ws = [0.2, 0.5]
        rp = 3
        rs = 90
        N, Wn = scp.signal.buttord(wp, ws, rp, rs, False)
        return N, Wn

    @pytest.mark.xfail(reason="TODO: freqs")
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_analog(self, xp, scp):
        wp = 200
        ws = 600
        rp = 3
        rs = 60
        N, Wn = scp.signal.buttord(wp, ws, rp, rs, True)
        b, a = scp.signal.butter(N, Wn, 'lowpass', True)
        w, h = scp.signal.freqs(b, a)
        return w, h

#        assert_array_less(-rp, dB(h[w <= wp]))
#        assert_array_less(dB(h[ws <= w]), -rs)
#
#        assert_equal(N, 7)
#        assert_allclose(Wn, 2.0006785355671877e+02, rtol=1e-15)
#
#        n, Wn = buttord(1, 550/450, 1, 26, analog=True)
#        assert_equal(n, 19)
#        assert_allclose(Wn, 1.0361980524629517, rtol=1e-15)
#
#        assert_equal(buttord(1, 1.2, 1, 80, analog=True)[0], 55)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_fs_param(self, xp, scp):
        wp = [4410, 11025]
        ws = [2205, 13230]
        rp = 3
        rs = 80
        fs = 44100
        N, Wn = scp.signal.buttord(wp, ws, rp, rs, False, fs=fs)
        return N, Wn

    def test_invalid_input(self):
        with pytest.raises(ValueError) as exc_info:
            signal.buttord([20, 50], [14, 60], 3, 2)
        assert "gpass should be smaller than gstop" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.buttord([20, 50], [14, 60], -1, 2)
        assert "gpass should be larger than 0.0" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.buttord([20, 50], [14, 60], 1, -2)
        assert "gstop should be larger than 0.0" in str(exc_info.value)

    def test_runtime_warnings(self):
        with pytest.warns(RuntimeWarning, match=r'Order is zero'):
            signal.buttord(0.0, 1.0, 3, 60)

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1E-7)
    def test_ellip_butter(self, xp, scp):
        # The purpose of the test is to make sure the result of `ellipord`
        # differs from that of `buttord`. The values to compare to are
        # generated with scipy 1.9.1
        n, wn = scp.signal.buttord([0.1, 0.6], [0.2, 0.5], 3, 60)
        return n, wn


@testing.with_requires("scipy")
class TestCheb1ord:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_lowpass(self, xp, scp):
        wp = 0.2
        ws = 0.3
        rp = 3
        rs = 60
        N, Wn = scp.signal.cheb1ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_highpass(self, xp, scp):
        wp = 0.3
        ws = 0.2
        rp = 3
        rs = 70
        N, Wn = scp.signal.cheb1ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandpass(self, xp, scp):
        wp = [0.2, 0.5]
        ws = [0.1, 0.6]
        rp = 3
        rs = 80
        N, Wn = scp.signal.cheb1ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=3e-7)
    def test_bandstop(self, xp, scp):
        wp = [0.1, 0.6]
        ws = [0.2, 0.5]
        rp = 3
        rs = 90
        N, Wn = scp.signal.cheb1ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_analog(self, xp, scp):
        wp = 700
        ws = 100
        rp = 3
        rs = 70
        N, Wn = scp.signal.cheb1ord(wp, ws, rp, rs, True)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_fs_param(self, xp, scp):
        wp = 4800
        ws = 7200
        rp = 3
        rs = 60
        fs = 48000
        N, Wn = scp.signal.cheb1ord(wp, ws, rp, rs, False, fs=fs)
        return N, Wn

    def test_invalid_input(self):
        with pytest.raises(ValueError) as exc_info:
            signal.cheb1ord(0.2, 0.3, 3, 2)
        assert "gpass should be smaller than gstop" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.cheb1ord(0.2, 0.3, -1, 2)
        assert "gpass should be larger than 0.0" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.cheb1ord(0.2, 0.3, 1, -2)
        assert "gstop should be larger than 0.0" in str(exc_info.value)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_ellip_butter(self, xp, scp):
        # The purpose of the test is to make sure the result of `cheb1ord`
        # differs from that of `buttord`. The values to compare to are
        # generated with scipy 1.9.1

        n, wn = scp.signal.cheb1ord([0.1, 0.6], [0.2, 0.5], 3, 60)
        return n, wn


@testing.with_requires("scipy")
class TestCheb2ord:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_lowpass(self, xp, scp):
        wp = 0.2
        ws = 0.3
        rp = 3
        rs = 60
        N, Wn = scp.signal.cheb2ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_highpass(self, xp, scp):
        wp = 0.3
        ws = 0.2
        rp = 3
        rs = 70
        N, Wn = scp.signal.cheb2ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandpass(self, xp, scp):
        wp = [0.2, 0.5]
        ws = [0.1, 0.6]
        rp = 3
        rs = 80
        N, Wn = scp.signal.cheb2ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=3e-7)
    def test_bandstop(self, xp, scp):
        wp = [0.1, 0.6]
        ws = [0.2, 0.5]
        rp = 3
        rs = 90
        N, Wn = scp.signal.cheb2ord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_analog(self, xp, scp):
        wp = [20, 50]
        ws = [10, 60]
        rp = 3
        rs = 80
        N, Wn = scp.signal.cheb2ord(wp, ws, rp, rs, True)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_fs_param(self, xp, scp):
        wp = 150
        ws = 100
        rp = 3
        rs = 70
        fs = 1000
        N, Wn = scp.signal.cheb2ord(wp, ws, rp, rs, False, fs=fs)
        return N, Wn

    def test_invalid_input(self):
        with pytest.raises(ValueError) as exc_info:
            signal.cheb2ord([0.1, 0.6], [0.2, 0.5], 3, 2)
        assert "gpass should be smaller than gstop" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.cheb2ord([0.1, 0.6], [0.2, 0.5], -1, 2)
        assert "gpass should be larger than 0.0" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.cheb2ord([0.1, 0.6], [0.2, 0.5], 1, -2)
        assert "gstop should be larger than 0.0" in str(exc_info.value)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_ellip_butter(self, xp, scp):
        # The purpose of the test is to make sure the result of `cheb2ord`
        # differs from that of `buttord`. The values to compare to are
        # generated with scipy 1.9.1
        n, wn = scp.signal.cheb2ord([0.1, 0.6], [0.2, 0.5], 3, 60)
        return n, wn


@testing.with_requires("scipy")
class TestEllipord:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_lowpass(self, xp, scp):
        wp = 0.2
        ws = 0.3
        rp = 3
        rs = 60
        N, Wn = scp.signal.ellipord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    @testing.with_requires('scipy>=1.7')
    def test_lowpass_1000dB(self, xp, scp):
        # failed when ellipkm1 wasn't used in ellipord and ellipap
        wp = 0.2
        ws = 0.3
        rp = 3
        rs = 1000
        N, Wn = scp.signal.ellipord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_highpass(self, xp, scp):
        wp = 0.3
        ws = 0.2
        rp = 3
        rs = 70
        N, Wn = scp.signal.ellipord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bandpass(self, xp, scp):
        wp = xp.array([0.2, 0.5])
        ws = xp.array([0.1, 0.6])
        rp = 3
        rs = 80
        N, Wn = scp.signal.ellipord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=3E-7)
    def test_bandstop(self, xp, scp):
        wp = xp.array([0.1, 0.6])
        ws = xp.array([0.2, 0.5])
        rp = 3
        rs = 90
        N, Wn = scp.signal.ellipord(wp, ws, rp, rs, False)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_analog(self, xp, scp):
        wp = [1000, 6000]
        ws = [2000, 5000]
        rp = 3
        rs = 90
        N, Wn = scp.signal.ellipord(wp, ws, rp, rs, True)
        return N, Wn

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=3E-7)
    def test_fs_param(self, xp, scp):
        wp = [400, 2400]
        ws = [800, 2000]
        rp = 3
        rs = 90
        fs = 8000
        N, Wn = scp.signal.ellipord(wp, ws, rp, rs, False, fs=fs)
        return N, Wn

    def test_invalid_input(self):
        with pytest.raises(ValueError) as exc_info:
            signal.ellipord(0.2, 0.5, 3, 2)
        assert "gpass should be smaller than gstop" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.ellipord(0.2, 0.5, -1, 2)
        assert "gpass should be larger than 0.0" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            signal.ellipord(0.2, 0.5, 1, -2)
        assert "gstop should be larger than 0.0" in str(exc_info.value)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=3E-7)
    def test_ellip_butter(self, xp, scp):
        # The purpose of the test is to make sure the result of `ellipord`
        # differs from that of `buttord`. The values to compare to are
        # generated with scipy 1.9.1
        n, wn = scp.signal.ellipord([0.1, 0.6], [0.2, 0.5], 3, 60)
        return n, wn


@testing.with_requires("scipy")
class TestIIRDesign:

    def test_exceptions(self):
        with pytest.raises(ValueError, match="the same shape"):
            iirdesign(0.2, [0.1, 0.3], 1, 40)
        with pytest.raises(ValueError, match="the same shape"):
            iirdesign(cupy.array([[0.3, 0.6], [0.3, 0.6]]),
                      cupy.array([[0.4, 0.5], [0.4, 0.5]]), 1, 40)

        # discrete filter with non-positive frequency
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign(0, 0.5, 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign(-0.1, 0.5, 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign(0.1, 0, 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign(0.1, -0.5, 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0, 0.3], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([-0.1, 0.3], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, 0], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, -0.3], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, 0.3], [0, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, 0.3], [-0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, 0.3], [0.1, 0], 1, 40)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, 0.3], [0.1, -0.5], 1, 40)

        # analog filter with negative frequency
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign(-0.1, 0.5, 1, 40, analog=True)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign(0.1, -0.5, 1, 40, analog=True)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([-0.1, 0.3], [0.1, 0.5], 1, 40, analog=True)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, -0.3], [0.1, 0.5], 1, 40, analog=True)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, 0.3], [-0.1, 0.5], 1, 40, analog=True)
        with pytest.raises(ValueError, match="must be greater than 0"):
            iirdesign([0.1, 0.3], [0.1, -0.5], 1, 40, analog=True)

        # discrete filter with fs=None, freq > 1
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign(1, 0.5, 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign(1.1, 0.5, 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign(0.1, 1, 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign(0.1, 1.5, 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([1, 0.3], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([1.1, 0.3], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([0.1, 1], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([0.1, 1.1], [0.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([0.1, 0.3], [1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([0.1, 0.3], [1.1, 0.5], 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([0.1, 0.3], [0.1, 1], 1, 40)
        with pytest.raises(ValueError, match="must be less than 1"):
            iirdesign([0.1, 0.3], [0.1, 1.5], 1, 40)

        # discrete filter with fs>2, wp, ws < fs/2 must pass
        iirdesign(100, 500, 1, 40, fs=2000)
        iirdesign(500, 100, 1, 40, fs=2000)
        iirdesign([200, 400], [100, 500], 1, 40, fs=2000)
        iirdesign([100, 500], [200, 400], 1, 40, fs=2000)

        # discrete filter with fs>2, freq > fs/2: this must raise
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign(1000, 400, 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign(1100, 500, 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign(100, 1000, 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign(100, 1100, 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([1000, 400], [100, 500], 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([1100, 400], [100, 500], 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([200, 1000], [100, 500], 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([200, 1100], [100, 500], 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([200, 400], [1000, 500], 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([200, 400], [1100, 500], 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([200, 400], [100, 1000], 1, 40, fs=2000)
        with pytest.raises(ValueError, match="must be less than fs/2"):
            iirdesign([200, 400], [100, 1100], 1, 40, fs=2000)

        with pytest.raises(ValueError, match="strictly inside stopband"):
            iirdesign([0.1, 0.4], [0.5, 0.6], 1, 40)
        with pytest.raises(ValueError, match="strictly inside stopband"):
            iirdesign([0.5, 0.6], [0.1, 0.4], 1, 40)
        with pytest.raises(ValueError, match="strictly inside stopband"):
            iirdesign([0.3, 0.6], [0.4, 0.7], 1, 40)
        with pytest.raises(ValueError, match="strictly inside stopband"):
            iirdesign([0.4, 0.7], [0.3, 0.6], 1, 40)


@testing.with_requires("scipy")
class TestIIRNotch:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_ba_output(self, xp, scp):
        # Compare coefficients with Matlab ones
        # for the equivalent input:
        b, a = scp.signal.iirnotch(0.06, 30)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_frequency_response(self, xp, scp):
        # Get filter coefficients
        b, a = scp.signal.iirnotch(0.3, 30)
        return b, a

    def test_errors(self):
        # Exception should be raised if w0 > 1 or w0 <0
        assert_raises(ValueError, signal.iirnotch, w0=2, Q=30)
        assert_raises(ValueError, signal.iirnotch, w0=-1, Q=30)

        # Exception should be raised if any of the parameters
        # are not float (or cannot be converted to one)
        assert_raises(ValueError, signal.iirnotch, w0="blabla", Q=30)
        assert_raises(TypeError, signal.iirnotch, w0=-1, Q=[1, 2, 3])

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_fs_param(self, xp, scp):
        # Get filter coefficients
        b, a = scp.signal.iirnotch(1500, 30, fs=10000)
        return b, a


@testing.with_requires("scipy")
class TestIIRPeak:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_ba_output(self, xp, scp):
        # Compare coefficients with Matlab ones
        # for the equivalent input:
        b, a = scp.signal.iirpeak(0.06, 30)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_frequency_response(self, xp, scp):
        # Get filter coefficients
        b, a = scp.signal.iirpeak(0.3, 30)
        return b, a

    def test_errors(self):
        # Exception should be raised if w0 > 1 or w0 <0
        assert_raises(ValueError, signal.iirpeak, w0=2, Q=30)
        assert_raises(ValueError, signal.iirpeak, w0=-1, Q=30)

        # Exception should be raised if any of the parameters
        # are not float (or cannot be converted to one)
        assert_raises(ValueError, signal.iirpeak, w0="blabla", Q=30)
        assert_raises(TypeError, signal.iirpeak, w0=-1, Q=[1, 2, 3])

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_fs_param(self, xp, scp):
        # Get filter coefficients
        b, a = scp.signal.iirpeak(1200, 30, fs=8000)
        return b, a


@testing.with_requires("scipy")
class TestIIRComb:
    # Test erroneous input cases
    def test_invalid_input(self):
        # w0 is <= 0 or >= fs / 2
        fs = 1000
        for args in [(-fs, 30), (0, 35), (fs / 2, 40), (fs, 35)]:
            with pytest.raises(ValueError, match='w0 must be between '):
                signal.iircomb(*args, fs=fs)

        # fs is not divisible by w0
        for args in [(120, 30), (157, 35)]:
            with pytest.raises(ValueError, match='fs must be divisible '):
                signal.iircomb(*args, fs=fs)

        with pytest.raises(ValueError, match='fs must be divisible '):
            signal.iircomb(w0=49.999/int(44100/2), Q=30)

        with pytest.raises(ValueError, match='fs must be divisible '):
            signal.iircomb(w0=49.999, Q=30, fs=44100)

        # Filter type is not notch or peak
        for args in [(0.2, 30, 'natch'), (0.5, 35, 'comb')]:
            with pytest.raises(ValueError, match='ftype must be '):
                signal.iircomb(*args)

    # Verify that the filter's frequency response contains a
    # notch at the cutoff frequency
    @testing.numpy_cupy_allclose(scipy_name="scp")
    @pytest.mark.parametrize('ftype', ('notch', 'peak'))
    def test_frequency_response(self, ftype, xp, scp):
        # Create a notching or peaking comb filter at 1000 Hz
        b, a = scp.signal.iircomb(1000, 30, ftype=ftype, fs=10000)
        return b, a

    # Verify pass_zero parameter
    @testing.with_requires("scipy>=1.9.0")
    @testing.numpy_cupy_allclose(scipy_name="scp")
    @pytest.mark.parametrize('ftype,pass_zero,peak,notch',
                             [('peak', True, 123.45, 61.725),
                              ('peak', False, 61.725, 123.45),
                              ('peak', None, 61.725, 123.45),
                              ('notch', None, 61.725, 123.45),
                              ('notch', True, 123.45, 61.725),
                              ('notch', False, 61.725, 123.45)])
    def test_pass_zero(self, ftype, pass_zero, peak, notch, xp, scp):
        # Create a notching or peaking comb filter
        b, a = scp.signal.iircomb(
            123.45, 30, ftype=ftype, fs=1234.5, pass_zero=pass_zero)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_iir_symmetry(self, xp, scp):
        b, a = scp.signal.iircomb(400, 30, fs=24000)
        return b, a

    # Verify filter coefficients with MATLAB's iircomb function
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_ba_output(self, xp, scp):
        b, a = scp.signal.iircomb(60, 35, ftype='notch', fs=600)
        return b, a

    # Verify filter coefficients with MATLAB's iircomb function
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_ba_output_2(self, xp, scp):
        b, a = scp.signal.iircomb(60, 35, ftype='peak', fs=600)
        return b, a

    # Verify that https://github.com/scipy/scipy/issues/14043 is fixed
    @testing.with_requires('scipy>=1.9.0')
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_nearest_divisor(self, xp, scp):
        # Create a notching comb filter
        b, a = scp.signal.iircomb(50/int(44100/2), 50.0, ftype='notch')
        return b, a


@testing.with_requires("scipy")
class TestZpk2Tf:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_identity(self, xp, scp):
        """Test the identity transfer function."""
        z = xp.array([])
        p = xp.array([])
        k = 1.
        b, a = scp.signal.zpk2tf(z, p, k)
        return b, a



import operator
import warnings

import numpy

import cupy
from cupy import _core


# rename builtin range for use in functions that take a range argument
_range = range


# TODO(unno): use searchsorted
_histogram_kernel = _core.ElementwiseKernel(
    'S x, raw T bins, int32 n_bins',
    'raw U y',
    '''
    if (x < bins[0] or bins[n_bins - 1] < x) {
        return;
    }
    int high = n_bins - 1;
    int low = 0;

    while (high - low > 1) {
        int mid = (high + low) / 2;
        if (bins[mid] <= x) {
            low = mid;
        } else {
            high = mid;
        }
    }
    atomicAdd(&y[low], U(1));
    ''',
    'cupy_histogram_kernel')


_weighted_histogram_kernel = _core.ElementwiseKernel(
    'S x, raw T bins, int32 n_bins, raw W weights',
    'raw Y y',
    '''
    if (x < bins[0] or bins[n_bins - 1] < x) {
        return;
    }
    int high = n_bins - 1;
    int low = 0;

    while (high - low > 1) {
        int mid = (high + low) / 2;
        if (bins[mid] <= x) {
            low = mid;
        } else {
            high = mid;
        }
    }
    atomicAdd(&y[low], (Y)weights[i]);
    ''',
    'cupy_weighted_histogram_kernel')


def _ravel_and_check_weights(a, weights):
    """ Check a and weights have matching shapes, and ravel both """

    # Ensure that the array is a "subtractable" dtype
    if a.dtype == cupy.bool_:
        warnings.warn('Converting input from {} to {} for compatibility.'
                      .format(a.dtype, cupy.uint8),
                      RuntimeWarning, stacklevel=3)
        a = a.astype(cupy.uint8)

    if weights is not None:
        if not isinstance(weights, cupy.ndarray):
            raise ValueError('weights must be a cupy.ndarray')
        if weights.shape != a.shape:
            raise ValueError(
                'weights should have the same shape as a.')
        weights = weights.ravel()
    a = a.ravel()
    return a, weights


def _get_outer_edges(a, range):
    """
    Determine the outer bin edges to use, from either the data or the range
    argument
    """
    if range is not None:
        first_edge, last_edge = range
        if first_edge > last_edge:
            raise ValueError(
                'max must be larger than min in range parameter.')
        if not (numpy.isfinite(first_edge) and numpy.isfinite(last_edge)):
            raise ValueError(
                'supplied range of [{}, {}] is not finite'.format(
                    first_edge, last_edge))
    elif a.size == 0:
        first_edge = 0.0
        last_edge = 1.0
    else:
        first_edge = float(a.min())
        last_edge = float(a.max())
        if not (cupy.isfinite(first_edge) and cupy.isfinite(last_edge)):
            raise ValueError(
                'autodetected range of [{}, {}] is not finite'.format(
                    first_edge, last_edge))

    # expand empty range to avoid divide by zero
    if first_edge == last_edge:
        first_edge = first_edge - 0.5
        last_edge = last_edge + 0.5

    return first_edge, last_edge


def _get_bin_edges(a, bins, range):
    """
    Computes the bins used internally by `histogram`.

    Args:
        a (ndarray): Ravelled data array
        bins (int or ndarray): Forwarded argument from `histogram`.
        range (None or tuple): Forwarded argument from `histogram`.

    Returns:
        bin_edges (ndarray): Array of bin edges
    """
    # parse the overloaded bins argument
    n_equal_bins = None
    bin_edges = None

    if isinstance(bins, str):
        raise NotImplementedError(
            'only integer and array bins are implemented')
    elif isinstance(bins, cupy.ndarray) or numpy.ndim(bins) == 1:
        # TODO(okuta): After #3060 is merged, `if cupy.ndim(bins) == 1:`.
        if isinstance(bins, cupy.ndarray):
            bin_edges = bins
        else:
            bin_edges = numpy.asarray(bins)

        if (bin_edges[:-1] > bin_edges[1:]).any():  # synchronize! when CuPy
            raise ValueError(
                '`bins` must increase monotonically, when an array')
        if isinstance(bin_edges, numpy.ndarray):
            bin_edges = cupy.asarray(bin_edges)
    elif numpy.ndim(bins) == 0:
        try:
            n_equal_bins = operator.index(bins)
        except TypeError:
            raise TypeError(
                '`bins` must be an integer, a string, or an array')
        if n_equal_bins < 1:
            raise ValueError('`bins` must be positive, when an integer')

        first_edge, last_edge = _get_outer_edges(a, range)
    else:
        raise ValueError('`bins` must be 1d, when an array')

    if n_equal_bins is not None:
        # numpy's gh-10322 means that type resolution rules are dependent on
        # array shapes. To avoid this causing problems, we pick a type now and
        # stick with it throughout.
        bin_type = cupy.result_type(first_edge, last_edge, a)
        if cupy.issubdtype(bin_type, cupy.integer):
            bin_type = cupy.result_type(bin_type, float)

        # bin edges must be computed
        bin_edges = cupy.linspace(
            first_edge, last_edge, n_equal_bins + 1,
            endpoint=True, dtype=bin_type)
    return bin_edges


def histogram(x, bins=10, range=None, weights=None, density=False):
    """Computes the histogram of a set of data.

    Args:
        x (cupy.ndarray): Input array.
        bins (int or cupy.ndarray): If ``bins`` is an int, it represents the
            number of bins. If ``bins`` is an :class:`~cupy.ndarray`, it
            represents a bin edges.
        range (2-tuple of float, optional): The lower and upper range of the
            bins.  If not provided, range is simply ``(x.min(), x.max())``.
            Values outside the range are ignored. The first element of the
            range must be less than or equal to the second. `range` affects the
            automatic bin computation as well. While bin width is computed to
            be optimal based on the actual data within `range`, the bin count
            will fill the entire range including portions containing no data.
        density (bool, optional): If False, the default, returns the number of
            samples in each bin. If True, returns the probability *density*
            function at the bin, ``bin_count / sample_count / bin_volume``.
        weights (cupy.ndarray, optional): An array of weights, of the same
            shape as `x`.  Each value in `x` only contributes its associated
            weight towards the bin count (instead of 1).
    Returns:
        tuple: ``(hist, bin_edges)`` where ``hist`` is a :class:`cupy.ndarray`
        storing the values of the histogram, and ``bin_edges`` is a
        :class:`cupy.ndarray` storing the bin edges.

    .. warning::

        This function may synchronize the device.

    .. seealso:: :func:`numpy.histogram`
    """

    if x.dtype.kind == 'c':
        # TODO(unno): comparison between complex numbers is not implemented
        raise NotImplementedError('complex number is not supported')

    if not isinstance(x, cupy.ndarray):
        raise ValueError('x must be a cupy.ndarray')

    x, weights = _ravel_and_check_weights(x, weights)
    bin_edges = _get_bin_edges(x, bins, range)

    if weights is None:
        y = cupy.zeros(bin_edges.size - 1, dtype=cupy.int64)
        # TODO(leofang): we temporarily remove CUB histogram support for now,
        # see cupy/cupy#7698. When it's ready, revert the commit that checked
        # in this comment to restore the support.
        _histogram_kernel(x, bin_edges, bin_edges.size, y)
    else:
        simple_weights = (
            cupy.can_cast(weights.dtype, cupy.float64) or
            cupy.can_cast(weights.dtype, cupy.complex128)
        )
        if not simple_weights:
            # object dtype such as Decimal are supported in NumPy, but not here
            raise NotImplementedError(
                'only weights with dtype that can be cast to float or complex '
                'are supported')
        if weights.dtype.kind == 'c':
            y = cupy.zeros(bin_edges.size - 1, dtype=cupy.complex128)
            _weighted_histogram_kernel(
                x, bin_edges, bin_edges.size, weights.real, y.real)
            _weighted_histogram_kernel(
                x, bin_edges, bin_edges.size, weights.imag, y.imag)
        else:
            if weights.dtype.kind in 'bui':
                y = cupy.zeros(bin_edges.size - 1, dtype=int)
            else:
                y = cupy.zeros(bin_edges.size - 1, dtype=cupy.float64)
            _weighted_histogram_kernel(
                x, bin_edges, bin_edges.size, weights, y)

    if density:
        db = cupy.array(cupy.diff(bin_edges), cupy.float64)
        return y / db / y.sum(), bin_edges
    return y, bin_edges


def histogramdd(sample, bins=10, range=None, weights=None, density=False):
    """Compute the multidimensional histogram of some data.

    Args:
        sample (cupy.ndarray): The data to be histogrammed. (N, D) or (D, N)
            array

            Note the unusual interpretation of sample when an array_like:

            * When an array, each row is a coordinate in a D-dimensional
              space - such as ``histogramdd(cupy.array([p1, p2, p3]))``.
            * When an array_like, each element is the list of values for single
              coordinate - such as ``histogramdd((X, Y, Z))``.

            The first form should be preferred.
        bins (int or tuple of int or cupy.ndarray): The bin specification:

            * A sequence of arrays describing the monotonically increasing bin
              edges along each dimension.
            * The number of bins for each dimension (nx, ny, ... =bins)
            * The number of bins for all dimensions (nx=ny=...=bins).
        range (sequence, optional): A sequence of length D, each an optional
            (lower, upper) tuple giving the outer bin edges to be used if the
            edges are not given explicitly in `bins`. An entry of None in the
            sequence results in the minimum and maximum values being used for
            the corresponding dimension. The default, None, is equivalent to
            passing a tuple of D None values.
        weights (cupy.ndarray): An array of values `w_i` weighing each sample
            `(x_i, y_i, z_i, ...)`. The values of the returned histogram are
            equal to the sum of the weights belonging to the samples falling
            into each bin.
        density (bool, optional): If False, the default, returns the number of
            samples in each bin. If True, returns the probability *density*
            function at the bin, ``bin_count / sample_count / bin_volume``.

    Returns:
        tuple:
        H (cupy.ndarray):
            The multidimensional histogram of sample x. See
            normed and weights for the different possible semantics.
        edges (list of cupy.ndarray):
            A list of D arrays describing the bin
            edges for each dimension.

    .. warning::

        This function may synchronize the device.

    .. seealso:: :func:`numpy.histogramdd`
    """
    if isinstance(sample, cupy.ndarray):
        # Sample is an ND-array.
        if sample.ndim == 1:
            sample = sample[:, cupy.newaxis]
        nsamples, ndim = sample.shape
    else:
        sample = cupy.stack(sample, axis=-1)
        nsamples, ndim = sample.shape

    nbin = numpy.empty(ndim, int)
    edges = ndim * [None]
    dedges = ndim * [None]
    if weights is not None:
        weights = cupy.asarray(weights)

    try:
        nbins = len(bins)
        if nbins != ndim:
            raise ValueError(
                'The dimension of bins must be equal to the dimension of the '
                ' sample x.'
            )
    except TypeError:
        # bins is an integer
        bins = ndim * [bins]

    # normalize the range argument
    if range is None:
        range = (None,) * ndim
    elif len(range) != ndim:
        raise ValueError('range argument must have one entry per dimension')

    # Create edge arrays
    for i in _range(ndim):
        if cupy.ndim(bins[i]) == 0:
            if bins[i] < 1:
                raise ValueError(
                    '`bins[{}]` must be positive, when an integer'.format(i)
                )
            smin, smax = _get_outer_edges(sample[:, i], range[i])
            num = int(bins[i] + 1)  # synchronize!
            edges[i] = cupy.linspace(smin, smax, num)
        elif cupy.ndim(bins[i]) == 1:
            if not isinstance(bins[i], cupy.ndarray):
                raise ValueError('array-like bins not supported')
            edges[i] = bins[i]
            if (edges[i][:-1] > edges[i][1:]).any():  # synchronize!
                raise ValueError(
                    '`bins[{}]` must be monotonically increasing, when an '
                    'array'.format(i)
                )
        else:
            raise ValueError(
                '`bins[{}]` must be a scalar or 1d array'.format(i)
            )

        nbin[i] = len(edges[i]) + 1  # includes an outlier on each end
        dedges[i] = cupy.diff(edges[i])

    # Compute the bin number each sample falls into.
    ncount = tuple(
        # avoid cupy.digitize to work around NumPy issue gh-11022
        cupy.searchsorted(edges[i], sample[:, i], side='right')
        for i in _range(ndim)
    )

    # Using digitize, values that fall on an edge are put in the right bin.
    # For the rightmost bin, we want values equal to the right edge to be
    # counted in the last bin, and not as an outlier.
    for i in _range(ndim):
        # Find which points are on the rightmost edge.
        on_edge = sample[:, i] == edges[i][-1]
        # Shift these points one bin to the left.
        ncount[i][on_edge] -= 1

    # Compute the sample indices in the flattened histogram matrix.
    # This raises an error if the array is too large.
    xy = cupy.ravel_multi_index(ncount, nbin)

    # Compute the number of repetitions in xy and assign it to the
    # flattened histmat.
    hist = cupy.bincount(xy, weights, minlength=numpy.prod(nbin))

    # Shape into a proper matrix
    hist = hist.reshape(nbin)

    # This preserves the (bad) behavior observed in NumPy gh-7845, for now.
    hist = hist.astype(float)  # Note: NumPy uses casting='safe' here too

    # Remove outliers (indices 0 and -1 for each dimension).
    core = ndim * (slice(1, -1),)
    hist = hist[core]

    if density:
        # calculate the probability density function
        s = hist.sum()
        for i in _range(ndim):
            shape = [1] * ndim
            shape[i] = nbin[i] - 2
            hist = hist / dedges[i].reshape(shape)
        hist /= s

    if any(hist.shape != numpy.asarray(nbin) - 2):
        raise RuntimeError('Internal Shape Error')
    return hist, edges


def histogram2d(x, y, bins=10, range=None, weights=None, density=None):
    """Compute the bi-dimensional histogram of two data samples.

    Args:
        x (cupy.ndarray): The first array of samples to be histogrammed.
        y (cupy.ndarray): The second array of samples to be histogrammed.
        bins (int or tuple of int or cupy.ndarray): The bin specification:

            * A sequence of arrays describing the monotonically increasing bin
              edges along each dimension.
            * The number of bins for each dimension (nx, ny)
            * The number of bins for all dimensions (nx=ny=bins).
        range (sequence, optional): A sequence of length two, each an optional
            (lower, upper) tuple giving the outer bin edges to be used if the
            edges are not given explicitly in `bins`. An entry of None in the
            sequence results in the minimum and maximum values being used for
            the corresponding dimension. The default, None, is equivalent to
            passing a tuple of two None values.
        weights (cupy.ndarray): An array of values `w_i` weighing each sample
            `(x_i, y_i)`. The values of the returned histogram are equal to the
            sum of the weights belonging to the samples falling into each bin.
        density (bool, optional): If False, the default, returns the number of
            samples in each bin. If True, returns the probability *density*
            function at the bin, ``bin_count / sample_count / bin_volume``.

    Returns:
        tuple:
        H (cupy.ndarray):
            The multidimensional histogram of sample x. See
            normed and weights for the different possible semantics.
        edges0 (tuple of cupy.ndarray):
            A list of D arrays describing the bin
            edges for the first dimension.
        edges1 (tuple of cupy.ndarray):
            A list of D arrays describing the bin
            edges for the second dimension.

    .. warning::

        This function may synchronize the device.

    .. seealso:: :func:`numpy.histogram2d`
    """
    try:
        n = len(bins)
    except TypeError:
        n = 1

    if n != 1 and n != 2:
        if isinstance(bins, cupy.ndarray):
            xedges = yedges = bins
            bins = [xedges, yedges]
        else:
            raise ValueError('array-like bins not supported in CuPy')

    hist, edges = histogramdd([x, y], bins, range, weights, density)
    return hist, edges[0], edges[1]


_bincount_kernel = _core.ElementwiseKernel(
    'S x', 'raw U bin',
    'atomicAdd(&bin[x], U(1))',
    'cupy_bincount_kernel')


_bincount_with_weight_kernel = _core.ElementwiseKernel(
    'S x, T w', 'raw U bin',
    'atomicAdd(&bin[x], w)',
    'cupy_bincount_with_weight_kernel')


def bincount(x, weights=None, minlength=None):
    """Count number of occurrences of each value in array of non-negative ints.

    Args:
        x (cupy.ndarray): Input array.
        weights (cupy.ndarray): Weights array which has the same shape as
            ``x``.
        minlength (int): A minimum number of bins for the output array.

    Returns:
        cupy.ndarray: The result of binning the input array. The length of
        output is equal to ``max(cupy.max(x) + 1, minlength)``.

    .. warning::

        This function may synchronize the device.

    .. seealso:: :func:`numpy.bincount`

    """
    if x.ndim > 1:
        raise ValueError('object too deep for desired array')
    if x.ndim < 1:
        raise ValueError('object of too small depth for desired array')
    if x.dtype.kind == 'f':
        raise TypeError('x must be int array')
    if (x < 0).any():  # synchronize!
        raise ValueError('The first argument of bincount must be non-negative')
    if weights is not None and x.shape != weights.shape:
        raise ValueError('The weights and list don\'t have the same length.')
    if minlength is not None:
        minlength = int(minlength)
        if minlength < 0:
            raise ValueError('minlength must be non-negative')

    size = int(cupy.max(x)) + 1  # synchronize!
    if minlength is not None:
        size = max(size, minlength)

    if weights is None:
        b = cupy.zeros((size,), dtype=numpy.intp)
        # TODO(leofang): we temporarily remove CUB histogram support for now,
        # see cupy/cupy#7698. When it's ready, revert the commit that checked
        # in this comment to restore the support.
        _bincount_kernel(x, b)
    else:
        b = cupy.zeros((size,), dtype=numpy.float64)
        _bincount_with_weight_kernel(x, weights, b)

    return b


def digitize(x, bins, right=False):
    """Finds the indices of the bins to which each value in input array belongs.

    .. note::

        In order to avoid device synchronization, digitize does not raise
        an exception when the array is not monotonic

    Args:
        x (cupy.ndarray): Input array.
        bins (cupy.ndarray): Array of bins.
            It has to be 1-dimensional and monotonic increasing or decreasing.
        right (bool):
            Indicates whether the intervals include the right or the left bin
            edge.

    Returns:
        cupy.ndarray: Output array of indices, of same shape as ``x``.

    .. seealso:: :func:`numpy.digitize`
    """  # NOQA
    # This is for NumPy compat, although it works fine
    if x.dtype.kind == 'c':
        raise TypeError('x may not be complex')

    if bins.ndim > 1:
        raise ValueError('object too deep for desired array')
    if bins.ndim < 1:
        raise ValueError('object of too small depth for desired array')

    # As the order of the arguments are reversed, the side must be too.
    side = 'left' if right else 'right'
    return cupy._sorting.search._searchsorted(bins, x, side, None, False)



import unittest

import cupy
from cupy import testing
import cupyx.scipy.special  # NOQA


@testing.with_requires('scipy<1.16')
class TestSpecial:

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(rtol=1e-5, scipy_name='scp')
    def check_unary(self, name, xp, scp, dtype):
        import scipy.special  # NOQA

        a = testing.shaped_arange((2, 3), xp, dtype)
        return getattr(scp.special, name)(a)

    def test_j0(self):
        self.check_unary('j0')

    def test_j1(self):
        self.check_unary('j1')

    def test_y0(self):
        self.check_unary('y0')

    def test_y1(self):
        self.check_unary('y1')

    def test_i0(self):
        self.check_unary('i0')

    def test_i0e(self):
        self.check_unary('i0e')

    def test_i1(self):
        self.check_unary('i1')

    def test_i1e(self):
        self.check_unary('i1e')

    def test_k0(self):
        self.check_unary('k0')

    def test_k0e(self):
        self.check_unary('k0e')

    def test_k1(self):
        self.check_unary('k1')

    def test_k1e(self):
        self.check_unary('k1e')

    @testing.for_dtypes('iId', name='order_dtype')
    @testing.for_dtypes('efd')
    @testing.numpy_cupy_allclose(atol=1e-12, scipy_name='scp')
    def test_yn(self, xp, scp, dtype, order_dtype):
        import scipy.special  # NOQA

        n = xp.arange(0, 10, dtype=order_dtype)
        a = xp.linspace(-10, 10, 100, dtype=dtype)
        return scp.special.yn(n[:, xp.newaxis], a[xp.newaxis, :])


@testing.with_requires('scipy<1.16')
class TestFusionSpecial(unittest.TestCase):

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(rtol=1e-5, scipy_name='scp')
    def check_unary(self, name, xp, scp, dtype):
        import scipy.special  # NOQA

        a = testing.shaped_arange((2, 3), xp, dtype)

        @cupy.fuse()
        def f(x):
            return getattr(scp.special, name)(x)

        return f(a)

    def test_j0(self):
        self.check_unary('j0')

    def test_j1(self):
        self.check_unary('j1')

    def test_y0(self):
        self.check_unary('y0')

    def test_y1(self):
        self.check_unary('y1')

    def test_i0(self):
        self.check_unary('i0')

    def test_i0e(self):
        self.check_unary('i0e')

    def test_i1(self):
        self.check_unary('i1')

    def test_i1e(self):
        self.check_unary('i1e')

    def test_k0(self):
        self.check_unary('k0')

    def test_k0e(self):
        self.check_unary('k0e')

    def test_k1(self):
        self.check_unary('k1')

    def test_k1e(self):
        self.check_unary('k1e')

    @testing.with_requires("scipy<1.14")
    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(rtol=1e-5, scipy_name='scp')
    def test_chbevl_dependent_fusion(self, dtype, xp, scp):
        @cupy.fuse
        def fused(x):
            _k0 = scp.special.k0(x)
            _k0e = scp.special.k0e(x)
            _k1 = scp.special.k1(x)
            _k1e = scp.special.k1e(x)

            # rgamma depends on `chbevl` too
            # see rgamma definition for additional info
            _rgamma = scp.special.rgamma(x)

            return _k0 + _k0e + _k1 + _k1e + _rgamma

        a = xp.linspace(-10, 10, 50, dtype=dtype).reshape((2, -1))
        return fused(a)



import operator
import unittest

import numpy
import pytest

import cupy
from cupy._core import _routines_linalg as _linalg
from cupy import testing


@testing.parameterize(
    *testing.product({
        'shape_pair': [
            # dot test
            ((3, 2), (2, 4)),
            ((3, 0), (0, 4)),
            ((0, 2), (2, 4)),
            ((3, 2), (2, 0)),
            ((2,), (2, 4)),
            ((0,), (0, 4)),
            ((3, 2), (2,)),
            ((3, 0), (0,)),
            ((2,), (2,)),
            ((0,), (0,)),
            # matmul test
            ((5, 3, 2), (5, 2, 4)),
            ((0, 3, 2), (0, 2, 4)),
            ((5, 3, 2), (2, 4)),
            ((0, 3, 2), (2, 4)),
            ((3, 2), (5, 2, 4)),
            ((3, 2), (0, 2, 4)),
            ((5, 3, 2), (1, 2, 4)),
            ((0, 3, 2), (1, 2, 4)),
            ((1, 3, 2), (5, 2, 4)),
            ((1, 3, 2), (0, 2, 4)),
            ((5, 3, 2), (2,)),
            ((5, 3, 0), (0,)),
            ((2,), (5, 2, 4)),
            ((0,), (5, 0, 4)),
            ((2, 2, 3, 2), (2, 2, 2, 4)),
            ((5, 0, 3, 2), (5, 0, 2, 4)),
            ((6, 5, 3, 2), (2, 4)),
            ((5, 0, 3, 2), (2, 4)),
            ((3, 2), (6, 5, 2, 4)),
            ((3, 2), (5, 0, 2, 4)),
            ((1, 5, 3, 2), (6, 1, 2, 4)),
            ((1, 0, 3, 2), (6, 1, 2, 4)),
            ((6, 1, 3, 2), (1, 5, 2, 4)),
            ((6, 1, 3, 2), (1, 0, 2, 4)),
            ((6, 5, 3, 2), (2,)),
            ((6, 5, 3, 0), (0,)),
            ((2,), (6, 5, 2, 4)),
            ((0,), (6, 5, 0, 4)),
            ((1, 3, 3), (10, 1, 3, 1)),
        ],
    }))
class TestMatmul(unittest.TestCase):

    @testing.for_all_dtypes(name='dtype1')
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_operator_matmul(self, xp, dtype1, dtype2):
        x1 = testing.shaped_arange(self.shape_pair[0], xp, dtype1)
        x2 = testing.shaped_arange(self.shape_pair[1], xp, dtype2)
        return operator.matmul(x1, x2)

    @testing.for_all_dtypes(name='dtype1')
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_cupy_matmul(self, xp, dtype1, dtype2):
        x1 = testing.shaped_arange(self.shape_pair[0], xp, dtype1)
        x2 = testing.shaped_arange(self.shape_pair[1], xp, dtype2)
        return xp.matmul(x1, x2)


@testing.parameterize(
    *testing.product({
        'shape_pair': [
            # dot test
            ((2, 3), (3, 4), (2, 4)),
            # ((0,), (0,), (0,)),  # TODO: fix GUFunc bug?
            # matmul test
            ((5, 3, 2), (5, 2, 4), (5, 3, 4)),
            ((0, 3, 2), (0, 2, 4), (0, 3, 4)),
        ],
    }))
class TestMatmulOut(unittest.TestCase):

    @testing.for_all_dtypes(name='dtype1')
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_allclose(
        rtol=1e-3, atol=1e-3,  # required for uint8
        accept_error=TypeError)
    def test_cupy_matmul_noncontiguous(self, xp, dtype1, dtype2):
        x1 = testing.shaped_arange(self.shape_pair[0], xp, dtype1)
        x2 = testing.shaped_arange(self.shape_pair[1], xp, dtype2)
        out = xp.zeros(self.shape_pair[2], dtype1)[::-1]
        ret = xp.matmul(x1, x2, out=out)
        # TODO: Fix GUFunc bug
        # assert ret is out
        assert xp.allclose(ret, out)
        return ret

    @testing.for_all_dtypes(name='dtype1')
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_cupy_matmul_out_cast(self, xp, dtype1, dtype2):
        x1 = testing.shaped_arange(self.shape_pair[0], xp, dtype1)
        x2 = testing.shaped_arange(self.shape_pair[1], xp, dtype2)
        out = xp.zeros(self.shape_pair[2], bool)
        ret = xp.matmul(x1, x2, out=out, casting='unsafe')
        # TODO: Fix GUFunc bug
        # assert ret is out
        assert xp.allclose(ret, out)
        return ret


class TestMatmulOutOverlap:

    @pytest.mark.parametrize('shape', [
        (900, 900),
        (2, 600, 600),
    ])
    @testing.for_dtypes([numpy.int32, numpy.float64])
    @testing.numpy_cupy_allclose(rtol=1e-5, atol=1e-5)
    def test_overlap_both(self, xp, dtype, shape):
        a = xp.ones(shape, dtype)
        return xp.matmul(a, a, out=a)


class TestMatmulStrides:

    @testing.for_all_dtypes()
    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_relaxed_c_contiguous_input(self, xp, dtype):
        x1 = testing.shaped_arange((2, 2, 3), xp, dtype)[:, None, :, :]
        x2 = testing.shaped_arange((2, 1, 3, 1), xp, dtype)
        return x1 @ x2


@testing.parameterize(
    *testing.product({
        'shape_pair': [
            ((6, 5, 3, 2), (6, 5, 2, 4)),
            ((6, 5, 3, 2), (6, 1, 2, 4)),
            ((6, 5, 3, 2), (1, 5, 2, 4)),
            ((6, 5, 3, 2), (1, 1, 2, 4)),
            ((6, 1, 3, 2), (6, 5, 2, 4)),
            ((1, 5, 3, 2), (6, 5, 2, 4)),
            ((1, 1, 3, 2), (6, 5, 2, 4)),
            ((3, 2), (6, 5, 2, 4)),
            ((6, 5, 3, 2), (2, 4)),
            ((2,), (6, 5, 2, 4)),
            ((6, 5, 3, 2), (2,)),
        ],
    }))
class TestMatmulLarge(unittest.TestCase):

    # Avoid overflow
    skip_dtypes = {
        (numpy.int8, numpy.uint8),
        (numpy.int8, numpy.int16),
        (numpy.int8, numpy.float16),
        (numpy.uint8, numpy.uint8),
        (numpy.uint8, numpy.int16),
        (numpy.uint8, numpy.uint16),
        (numpy.int16, numpy.int16),
        (numpy.uint16, numpy.uint16),
    }

    @testing.for_all_dtypes(name='dtype1')
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_operator_matmul(self, xp, dtype1, dtype2):
        if ((dtype1, dtype2) in self.skip_dtypes or
                (dtype2, dtype1) in self.skip_dtypes):
            pytest.skip()
        x1 = testing.shaped_random(self.shape_pair[0], xp, dtype1)
        x2 = testing.shaped_random(self.shape_pair[1], xp, dtype2)
        return operator.matmul(x1, x2)

    @testing.for_all_dtypes(name='dtype1')
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_cupy_matmul(self, xp, dtype1, dtype2):
        if ((dtype1, dtype2) in self.skip_dtypes or
                (dtype2, dtype1) in self.skip_dtypes):
            pytest.skip()
        shape1, shape2 = self.shape_pair
        x1 = testing.shaped_random(shape1, xp, dtype1)
        x2 = testing.shaped_random(shape2, xp, dtype2)
        return xp.matmul(x1, x2)


@pytest.mark.parametrize('shape1,shape2', [
    ((256, 256, 3, 2), (256, 256, 2, 4)),
    ((256, 256, 3, 2), (2, 4)),
    ((3, 2), (256, 256, 2, 4))
])
class TestMatmulIntegralLargeBatch:

    @testing.for_int_dtypes(name='dtype')
    @testing.numpy_cupy_array_equal()
    def test_operator_matmul(self, xp, dtype, shape1, shape2):
        x1 = testing.shaped_random(shape1, xp, dtype)
        x2 = testing.shaped_random(shape2, xp, dtype)
        return operator.matmul(x1, x2)

    @testing.for_int_dtypes(name='dtype')
    @testing.numpy_cupy_array_equal()
    def test_cupy_matmul(self, xp, dtype, shape1, shape2):
        x1 = testing.shaped_random(shape1, xp, dtype)
        x2 = testing.shaped_random(shape2, xp, dtype)
        return xp.matmul(x1, x2)


class TestMatmulOverflow(unittest.TestCase):

    @testing.for_int_dtypes(name='dtype', no_bool=True)
    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_overflow(self, xp, dtype):
        value = numpy.iinfo(dtype).max
        a = xp.array([value - 10]).astype(dtype)
        b = xp.array([value - 10]).astype(dtype)
        return xp.matmul(a, b)


class _TestMatmulComputeTypes(unittest.TestCase):

    def setUp(self):
        self.old_compute_type = cupy._core.get_compute_type(self.dtype)
        cupy._core.set_compute_type(self.dtype, self.compute_type)

    def tearDown(self):
        cupy._core.set_compute_type(self.dtype, self.old_compute_type)

    def make_x1_x2(self, xp, shapes, dtypes):
        x1 = testing.shaped_random(shapes[0], xp, dtypes[0])
        x2 = testing.shaped_random(shapes[1], xp, dtypes[1])
        return x1, x2


@testing.parameterize(
    *testing.product({
        'compute_type': [
            _linalg.COMPUTE_TYPE_DEFAULT,
            _linalg.COMPUTE_TYPE_PEDANTIC,
        ],
        'shape_pair': [
            ((32, 64), (64, 96)),
            ((64, 96), (96, 32)),
            ((96, 32), (32, 64)),
        ],
    }))
class TestMatmulFp16ComputeTypes(_TestMatmulComputeTypes):
    dtype = numpy.float16

    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)
    def test_operator_matmul(self, xp):
        x1, x2 = self.make_x1_x2(xp, self.shape_pair, (self.dtype, self.dtype))
        return operator.matmul(x1, x2)

    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)
    def test_cupy_matmul(self, xp):
        x1, x2 = self.make_x1_x2(xp, self.shape_pair, (self.dtype, self.dtype))
        return xp.matmul(x1, x2)


@testing.parameterize(
    *testing.product({
        'compute_type': [
            _linalg.COMPUTE_TYPE_DEFAULT,
            _linalg.COMPUTE_TYPE_PEDANTIC,
            _linalg.COMPUTE_TYPE_TF32,
        ],
        'shape_pair': [
            ((100, 200), (200, 300)),
            ((200, 300), (300, 100)),
            ((300, 100), (100, 200)),
        ],
        'dtype_pair': [
            (numpy.float16, numpy.float32),
            (numpy.float32, numpy.float32),
            (numpy.float16, numpy.complex64),
            (numpy.float32, numpy.complex64),
            (numpy.complex64, numpy.complex64),
        ],
    }))
class TestMatmulFp32ComputeTypes(_TestMatmulComputeTypes):
    dtype = numpy.float32

    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)
    def test_operator_matmul(self, xp):
        x1, x2 = self.make_x1_x2(xp, self.shape_pair, self.dtype_pair)
        return operator.matmul(x1, x2)

    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)
    def test_cupy_matmul(self, xp):
        x1, x2 = self.make_x1_x2(xp, self.shape_pair, self.dtype_pair)
        return xp.matmul(x1, x2)


@testing.parameterize(
    *testing.product({
        'compute_type': [
            _linalg.COMPUTE_TYPE_DEFAULT,
            _linalg.COMPUTE_TYPE_PEDANTIC,
        ],
        'shape_pair': [
            ((100, 200), (200, 300)),
            ((200, 300), (300, 100)),
            ((300, 100), (100, 200)),
        ],
        'dtype_pair': [
            (numpy.float32, numpy.float64),
            (numpy.float64, numpy.float64),
            (numpy.float32, numpy.complex128),
            (numpy.float64, numpy.complex128),
            (numpy.complex64, numpy.complex128),
            (numpy.complex128, numpy.complex128),
        ],
    }))
class TestMatmulFp64ComputeTypes(_TestMatmulComputeTypes):
    dtype = numpy.float64

    @testing.numpy_cupy_allclose()
    def test_operator_matmul(self, xp):
        x1, x2 = self.make_x1_x2(xp, self.shape_pair, self.dtype_pair)
        return operator.matmul(x1, x2)

    @testing.numpy_cupy_allclose()
    def test_cupy_matmul(self, xp):
        x1, x2 = self.make_x1_x2(xp, self.shape_pair, self.dtype_pair)
        return xp.matmul(x1, x2)


@testing.parameterize(
    *testing.product({
        'shape_pair': [
            ((5, 3, 1), (3, 1, 4)),
            ((3, 2, 3), (3, 2, 4)),
            ((3, 2), ()),
            ((), (3, 2)),
            ((), ()),
            ((3, 2), (1,)),
            ((0, 2), (3, 0)),
            ((0, 1, 1), (2, 1, 1)),
        ],
    }))
class TestMatmulInvalidShape(unittest.TestCase):

    def test_invalid_shape(self):
        for xp in (numpy, cupy):
            shape1, shape2 = self.shape_pair
            x1 = testing.shaped_arange(shape1, xp, numpy.float32)
            x2 = testing.shaped_arange(shape2, xp, numpy.float32)
            with pytest.raises(ValueError):
                xp.matmul(x1, x2)


@testing.parameterize(
    *testing.product({
        'shapes_axes': [
            (((2, 5, 3, 2, 3, 4),  (3, 5, 1, 1, 1, 4), (5, 5, 2, 2, 3, 4)),
             [(1, 2), (0, 1), (0, 1)]),
            (((2, 5, 3, 2, 3, 4),  (2, 5, 3, 1, 4, 1), (3, 1, 2, 5, 3, 2)),
             [(-2, -1), (-2, -1), (0, 1)]),
            (((3, 2, 4, 4), (4, 4, 3, 2), (4, 4, 3, 3)),
             [(0, 1), (-1, -2), (-2, -1)]),
            (((3, 2, 4, 4), (2, 3, 4, 4), (4, 3, 3, 4)),
             [(0, 1), (0, 1), (1, 2)]),
        ],
    }))
class TestMatmulAxes(unittest.TestCase):

    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_cupy_matmul_axes(self, xp):
        x1 = testing.shaped_arange(self.shapes_axes[0][0], xp)
        x2 = testing.shaped_arange(self.shapes_axes[0][1], xp)
        return xp.matmul(x1, x2, axes=self.shapes_axes[1])

    @testing.numpy_cupy_allclose(rtol=1e-3, atol=1e-3)  # required for uint8
    def test_cupy_matmul_axes_out(self, xp):
        x1 = testing.shaped_arange(self.shapes_axes[0][0], xp)
        x2 = testing.shaped_arange(self.shapes_axes[0][1], xp)
        out = xp.zeros(self.shapes_axes[0][2])
        xp.matmul(x1, x2, axes=self.shapes_axes[1], out=out)
        return out


class TestMatmulDispatch(unittest.TestCase):

    def test_matmul_dispatch(self):
        x1 = testing.shaped_arange((2, 10, 5), cupy)
        x2 = testing.shaped_arange((10, 2, 5), cupy)
        o_np = numpy.matmul(x1, x2, axes=[(0, 1), (0, 1), (0, 1)])
        assert isinstance(o_np, cupy.ndarray)
        o_cp = cupy.matmul(x1, x2, axes=[(0, 1), (0, 1), (0, 1)])
        testing.assert_allclose(o_np, o_cp)



import unittest

import cupy.testing._parameterized

try:
    import pytest
    import _pytest
    _error = None
except ImportError as e:
    pytest = None  # type: ignore
    _pytest = None  # type: ignore
    _error = e


def is_available():
    return _error is None and hasattr(pytest, 'fixture')


def check_available(feature):
    if not is_available():
        raise RuntimeError('''\
cupy.testing: {} is not available.

Reason: {}: {}'''.format(feature, type(_error).__name__, _error))


if is_available():

    class _TestingParameterizeMixin:

        def __repr__(self):
            return '<{}  parameter: {}>'.format(
                super().__repr__(),
                self.__dict__,
            )

        @pytest.fixture(autouse=True)
        def _cupy_testing_parameterize(self, _cupy_testing_param):
            assert not self.__dict__, \
                'There should not be another hack with instance attribute.'
            self.__dict__.update(_cupy_testing_param)


def parameterize(*params, _ids=True):
    check_available('parameterize')
    if _ids:
        param_name = cupy.testing._parameterized._make_class_name
    else:
        def param_name(_, i, param):
            return str(i)

    # TODO(kataoka): Give better names (`id`).
    # For now, use legacy `_make_class_name` just for consistency. Here,
    # a generated name is `TestFoo::test_bar[_param_0_{...}]`, whereas
    # a legacy name is `TestFoo_param_0_{...}::test_bar
    params = [
        pytest.param(param, id=param_name('', i, param))
        for i, param in enumerate(params)
    ]

    def f(cls):
        assert not issubclass(cls, unittest.TestCase)
        if issubclass(cls, _TestingParameterizeMixin):
            raise RuntimeError('do not `@testing.parameterize` twice')
        module_name = cls.__module__
        cls = type(cls.__name__, (_TestingParameterizeMixin, cls), {})
        cls.__module__ = module_name
        cls = pytest.mark.parametrize('_cupy_testing_param', params)(cls)
        return cls
    return f



import math
import unittest

import numpy
import pytest

import cupy
from cupy import testing
import cupyx.scipy.stats  # NOQA
from cupyx.scipy import stats
from cupyx.scipy.stats import _distributions

try:
    import scipy.stats  # NOQA
except ImportError:
    pass


class TestEntropyBasic(unittest.TestCase):
    def test_entropy_positive(self):
        # See ticket SciPy's gh-497
        pk = cupy.asarray([0.5, 0.2, 0.3])
        qk = cupy.asarray([0.1, 0.25, 0.65])
        eself = stats.entropy(pk, pk)
        edouble = stats.entropy(pk, qk)
        assert 0.0 == eself
        assert edouble >= 0.0

    def test_entropy_base(self):
        pk = cupy.ones(16, float)
        s = stats.entropy(pk, base=2.0)
        assert abs(s - 4.0) < 1.0e-5

        qk = cupy.ones(16, float)
        qk[:8] = 2.0
        s = stats.entropy(pk, qk)
        s2 = stats.entropy(pk, qk, base=2.0)
        assert abs(s / s2 - math.log(2.0)) < 1.0e-5

    def test_entropy_zero(self):
        # Test for SciPy PR-479
        s = stats.entropy(cupy.asarray([0, 1, 2]))
        expected = 0.63651416829481278
        assert abs(float(s) - expected) < 1e-12

    def test_entropy_2d(self):
        pk = cupy.asarray([[0.1, 0.2], [0.6, 0.3], [0.3, 0.5]])
        qk = cupy.asarray([[0.2, 0.1], [0.3, 0.6], [0.5, 0.3]])
        testing.assert_array_almost_equal(
            stats.entropy(pk, qk), [0.1933259, 0.18609809]
        )

    def test_entropy_2d_zero(self):
        pk = cupy.asarray([[0.1, 0.2], [0.6, 0.3], [0.3, 0.5]])
        qk = cupy.asarray([[0.0, 0.1], [0.3, 0.6], [0.5, 0.3]])
        testing.assert_array_almost_equal(stats.entropy(pk, qk),
                                          [cupy.inf, 0.18609809])

        pk[0][0] = 0.0
        testing.assert_array_almost_equal(
            stats.entropy(pk, qk), [0.17403988, 0.18609809]
        )

    def test_entropy_base_2d_nondefault_axis(self):
        pk = cupy.asarray([[0.1, 0.2], [0.6, 0.3], [0.3, 0.5]])
        testing.assert_array_almost_equal(
            stats.entropy(pk, axis=1),
            cupy.asarray([0.63651417, 0.63651417, 0.66156324]),
        )

    def test_entropy_2d_nondefault_axis(self):
        pk = cupy.asarray([[0.1, 0.2], [0.6, 0.3], [0.3, 0.5]])
        qk = cupy.asarray([[0.2, 0.1], [0.3, 0.6], [0.5, 0.3]])
        testing.assert_array_almost_equal(
            stats.entropy(pk, qk, axis=1),
            cupy.asarray([0.231049, 0.231049, 0.127706]),
        )

    def test_entropy_raises_value_error(self):
        pk = cupy.asarray([[0.1, 0.2], [0.6, 0.3], [0.3, 0.5]])
        qk = cupy.asarray([[0.1, 0.2], [0.6, 0.3]])
        with pytest.raises(ValueError):
            stats.entropy(pk, qk)


@testing.parameterize(*(
    testing.product({
        'shape': [(64, ), (16, 15), (14, 4, 10)],
        'base': [None, 10],
        'axis': [None, 0, -1],
        'use_qk': [False, True],
        'normalize': [False, True],
    })
))
@testing.with_requires("scipy<1.14.0")
class TestEntropy(unittest.TestCase):

    def _entropy(self, xp, scp, dtype, shape, use_qk, base, axis, normalize):
        pk = testing.shaped_random(shape, xp, dtype=dtype)
        is_float16 = pk.dtype.char == 'e'
        if use_qk:
            qk = testing.shaped_random(shape, xp, dtype=dtype)
        else:
            qk = None

        if normalize and pk.dtype.kind != 'c':
            # if we don't normalize pk and qk, entropy will do it internally
            norm_axis = 0 if axis is None else axis
            pk = _distributions._normalize(pk, norm_axis)
            if qk is not None:
                qk = _distributions._normalize(qk, norm_axis)
        res = scp.stats.entropy(pk, qk=qk, base=base, axis=axis)

        float_type = xp.float32 if pk.dtype.char in 'ef' else xp.float64
        if res.ndim > 0:
            # verify expected dtype
            assert res.dtype == float_type

        # Cast back to the floating precision of the input so that the
        # correct rtol is used by numpy_cupy_allclose
        res = xp.asarray(res, xp.float16 if is_float16 else float_type)
        return res

    @testing.for_all_dtypes(no_complex=True)
    @testing.numpy_cupy_allclose(rtol={cupy.float16: 1e-3,
                                       cupy.float32: 1e-6,
                                       'default': 1e-15},
                                 scipy_name='scp')
    def test_entropy(self, xp, scp, dtype):
        return self._entropy(xp, scp, dtype, self.shape, self.use_qk,
                             self.base, self.axis, self.normalize)

    @testing.for_complex_dtypes()
    def test_entropy_complex(self, dtype):
        for xp, scp in zip([numpy, cupy], [scipy, cupyx.scipy]):
            with pytest.raises(TypeError):
                return self._entropy(xp, scp, dtype, self.shape, self.use_qk,
                                     self.base, self.axis, self.normalize)



# distutils: language = c++

"""Thin wrapper of cuTENSOR."""

cimport cython  # NOQA
from libc.stdint cimport int32_t, uint32_t, int64_t, uint64_t, intptr_t

from cupy_backends.cuda cimport stream as stream_module
from cupy_backends.cuda.api cimport runtime

cdef extern from '../../cupy_cutensor.h' nogil:
    ctypedef void* Handle_t 'cutensorHandle_t'
    ctypedef void* TensorDescriptor_t 'cutensorTensorDescriptor_t'
    ctypedef void* OperationDescriptor_t 'cutensorOperationDescriptor_t'
    ctypedef void* PlanPreference_t 'cutensorPlanPreference_t'
    ctypedef void* Plan_t 'cutensorPlan_t'

    ctypedef int Status_t 'cutensorStatus_t'
    ctypedef int Algo_t 'cutensorAlgo_t'
    ctypedef int JitMode_t 'cutensorJitMode_t'
    ctypedef int Operator_t 'cutensorOperator_t'
    ctypedef int WorksizePreference_t 'cutensorWorksizePreference_t'
    ctypedef int DataType_t 'cutensorDataType_t'
    ctypedef int ComputeType_t 'cutensorComputeType_t'

    ctypedef struct cutensorComputeDescriptor:
        pass
    # ctypedef cutensorComputeDescriptor* cutensorComputeDescriptor_t
    ctypedef cutensorComputeDescriptor* ComputeDescriptor_t 'cutensorComputeDescriptor_t'  # NOQA
    const ComputeDescriptor_t CUTENSOR_COMPUTE_DESC_16F
    const ComputeDescriptor_t CUTENSOR_COMPUTE_DESC_16BF
    const ComputeDescriptor_t CUTENSOR_COMPUTE_DESC_TF32
    const ComputeDescriptor_t CUTENSOR_COMPUTE_DESC_3XTF32
    const ComputeDescriptor_t CUTENSOR_COMPUTE_DESC_32F
    const ComputeDescriptor_t CUTENSOR_COMPUTE_DESC_64F

    #
    const char* cutensorGetErrorString(Status_t status)
    size_t cutensorGetVersion()
    size_t cutensorGetCudartVersion()

    # Handle creation and destruction
    Status_t cutensorCreate(Handle_t* handle)
    Status_t cutensorDestroy(Handle_t handle)

    # TensorDescriptor creation and destruction
    Status_t cutensorCreateTensorDescriptor(
        Handle_t handle,
        TensorDescriptor_t* desc,
        uint32_t numModes,
        int64_t* extent,
        int64_t* stride,
        DataType_t dataType,
        uint32_t alignmentRequirement)
    Status_t cutensorDestroyTensorDescriptor(TensorDescriptor_t desc)

    # PlanPreference creation and destruction
    Status_t cutensorCreatePlanPreference(
        Handle_t handle,
        PlanPreference_t* pref,
        int algo,
        int jitMode)
    Status_t cutensorDestroyPlanPreference(PlanPreference_t pref)

    Status_t cutensorEstimateWorkspaceSize(
        Handle_t handle,
        OperationDescriptor_t desc,
        PlanPreference_t planPref,
        WorksizePreference_t workspacePref,
        uint64_t* workspaceSizeEstimate)

    # Plan creation and destruction
    Status_t cutensorCreatePlan(
        Handle_t handle,
        Plan_t* plan,
        OperationDescriptor_t desc,
        PlanPreference_t pref,
        uint64_t workspaceSizeLimit)
    Status_t cutensorDestroyPlan(Plan_t plan)

    # cutensorElementwiseTrinary
    Status_t cutensorCreateElementwiseTrinary(
        Handle_t handle, OperationDescriptor_t* desc,
        TensorDescriptor_t descA, int32_t* modeA, Operator_t opA,
        TensorDescriptor_t descB, int32_t* modeB, Operator_t opB,
        TensorDescriptor_t descC, int32_t* modeC, Operator_t opC,
        TensorDescriptor_t descD, int32_t* modeD,
        Operator_t opAB, Operator_t opABC, ComputeDescriptor_t descCompute)

    Status_t cutensorElementwiseTrinaryExecute(
        Handle_t handle, Plan_t plan,
        void* alpha, void* A,
        void* beta, void* B,
        void* gamma, void* C, void* D, runtime.Stream stream)

    # cutensorElementwiseBinary
    Status_t cutensorCreateElementwiseBinary(
        Handle_t handle, OperationDescriptor_t* desc,
        TensorDescriptor_t descA, int32_t* modeA, Operator_t opA,
        TensorDescriptor_t descC, int32_t* modeC, Operator_t opC,
        TensorDescriptor_t descD, int32_t* modeD,
        Operator_t opAC, ComputeDescriptor_t descCompute)

    Status_t cutensorElementwiseBinaryExecute(
        Handle_t handle, Plan_t plan,
        void* alpha, void* A,
        void* gamma, void* C, void* D, runtime.Stream stream)

    # cutensorPermutation
    Status_t cutensorCreatePermutation(
        Handle_t handle, OperationDescriptor_t* desc,
        TensorDescriptor_t descA, int32_t* modeA, Operator_t opA,
        TensorDescriptor_t descB, int32_t* modeB,
        ComputeDescriptor_t descCompute)

    Status_t cutensorPermute(
        Handle_t handle, Plan_t plan,
        void* alpha, void* A,
        void* B, runtime.Stream stream)

    # cutensorContraction
    Status_t cutensorCreateContraction(
        Handle_t handle, OperationDescriptor_t* desc,
        TensorDescriptor_t descA, int32_t* modeA, Operator_t opA,
        TensorDescriptor_t descB, int32_t* modeB, Operator_t opB,
        TensorDescriptor_t descC, int32_t* modeC, Operator_t opC,
        TensorDescriptor_t descD, int32_t* modeD,
        ComputeDescriptor_t descCompute)

    Status_t cutensorContract(
        Handle_t handle, Plan_t plan,
        void* alpha, void* A, void* B,
        void* beta, void* C, void* D,
        void* workspace, uint64_t workspaceSize, runtime.Stream stream)

    # cutensorReduction
    Status_t cutensorCreateReduction(
        Handle_t handle, OperationDescriptor_t* desc,
        TensorDescriptor_t descA, int32_t* modeA, Operator_t opA,
        TensorDescriptor_t descC, int32_t* modeC, Operator_t opC,
        TensorDescriptor_t descD, int32_t* modeD,
        Operator_t opReduce, ComputeDescriptor_t descCompute)

    Status_t cutensorReduce(
        Handle_t handle, Plan_t plan,
        void* alpha, void* A,
        void* beta, void* C, void* D,
        void* workspace, uint64_t workspaceSize, runtime.Stream stream)

    #
    Status_t cutensorDestroyOperationDescriptor(OperationDescriptor_t desc)

    # build-time version
    int CUTENSOR_VERSION


available = True


###############################################################################
# Version information
###############################################################################

cpdef size_t get_version():
    return cutensorGetVersion()

cpdef size_t get_cudart_version():
    return cutensorGetCudartVersion()

###############################################################################
# Error handling
###############################################################################


class CuTensorError(RuntimeError):

    def __init__(self, int status):
        self.status = status
        msg = cutensorGetErrorString(<Status_t>status)
        super(CuTensorError, self).__init__(msg.decode())

    def __reduce__(self):
        return (type(self), (self.status,))


@cython.profile(False)
cdef inline check_status(int status):
    if status != STATUS_SUCCESS:
        raise CuTensorError(status)


###############################################################################
# cutensorHandle creation and destruction
###############################################################################

cpdef intptr_t create() except? 0:
    cdef Handle_t handle
    with nogil:
        status = cutensorCreate(&handle)
    check_status(status)
    return <intptr_t>handle

cpdef destroy(intptr_t handle):
    with nogil:
        status = cutensorDestroy(<Handle_t>handle)
    check_status(status)


###############################################################################
# cutensorTensorDescriptor: creation and destruction
###############################################################################

cpdef intptr_t createTensorDescriptor(
        intptr_t handle,
        uint32_t numModes,
        intptr_t extent,
        intptr_t stride,
        int dataType,
        uint32_t alignmentRequirement) except? 0:
    cdef TensorDescriptor_t desc
    with nogil:
        status = cutensorCreateTensorDescriptor(
            <Handle_t>handle, &desc,
            numModes, <int64_t*>extent, <int64_t*>stride,
            <DataType_t>dataType, alignmentRequirement)
    check_status(status)
    return <intptr_t>desc

cpdef destroyTensorDescriptor(intptr_t desc):
    with nogil:
        status = cutensorDestroyTensorDescriptor(<TensorDescriptor_t>desc)
    check_status(status)

###############################################################################
# cutensorPlanPreference: creation and destruction
###############################################################################

cpdef intptr_t createPlanPreference(
        intptr_t handle,
        int algo,
        int jitMode) except? 0:
    cdef PlanPreference_t pref
    with nogil:
        status = cutensorCreatePlanPreference(
            <Handle_t>handle, &pref, <Algo_t>algo, <JitMode_t>jitMode)
    check_status(status)
    return <intptr_t>pref

cpdef destroyPlanPreference(intptr_t pref):
    with nogil:
        status = cutensorDestroyPlanPreference(<PlanPreference_t>pref)
    check_status(status)

cpdef uint64_t estimateWorkspaceSize(
        intptr_t handle,
        intptr_t desc,
        intptr_t planPref,
        int workspacePref):
    cdef uint64_t workspaceSizeEstimate
    with nogil:
        status = cutensorEstimateWorkspaceSize(
            <Handle_t>handle, <OperationDescriptor_t>desc,
            <PlanPreference_t>planPref, <WorksizePreference_t>workspacePref,
            &workspaceSizeEstimate)
    check_status(status)
    return workspaceSizeEstimate

###############################################################################
# cutensorPlan: creation and destruction
###############################################################################

cpdef intptr_t createPlan(
        intptr_t handle,
        intptr_t desc,
        intptr_t pref,
        uint64_t workspaceSizeLimit) except? 0:
    cdef Plan_t plan
    with nogil:
        status = cutensorCreatePlan(
            <Handle_t>handle, &plan, <OperationDescriptor_t>desc,
            <PlanPreference_t>pref, workspaceSizeLimit)
    check_status(status)
    return <intptr_t>plan

cpdef destroyPlan(intptr_t plan):
    with nogil:
        status = cutensorDestroyPlan(<Plan_t>plan)
    check_status(status)

###############################################################################

cdef ComputeDescriptor_t getComputeDesc(int descCompute) nogil:
    if (descCompute == COMPUTE_16F):
        return CUTENSOR_COMPUTE_DESC_16F
    elif (descCompute == COMPUTE_16BF):
        return CUTENSOR_COMPUTE_DESC_16BF
    elif (descCompute == COMPUTE_TF32):
        return CUTENSOR_COMPUTE_DESC_TF32
    elif (descCompute == COMPUTE_3XTF32):
        return CUTENSOR_COMPUTE_DESC_3XTF32
    elif (descCompute == COMPUTE_32F):
        return CUTENSOR_COMPUTE_DESC_32F
    elif (descCompute == COMPUTE_64F):
        return CUTENSOR_COMPUTE_DESC_64F
    return CUTENSOR_COMPUTE_DESC_32F

###############################################################################
# cutensorElementwiseTrinary
###############################################################################

cpdef intptr_t createElementwiseTrinary(
        intptr_t handle,
        intptr_t descA, intptr_t modeA, int opA,
        intptr_t descB, intptr_t modeB, int opB,
        intptr_t descC, intptr_t modeC, int opC,
        intptr_t descD, intptr_t modeD,
        int opAB, int opABC, int descCompute) except? 0:
    cdef OperationDescriptor_t desc
    with nogil:
        status = cutensorCreateElementwiseTrinary(
            <Handle_t>handle, &desc,
            <TensorDescriptor_t>descA, <int32_t*>modeA, <Operator_t>opA,
            <TensorDescriptor_t>descB, <int32_t*>modeB, <Operator_t>opB,
            <TensorDescriptor_t>descC, <int32_t*>modeC, <Operator_t>opC,
            <TensorDescriptor_t>descD, <int32_t*>modeD,
            <Operator_t>opAB, <Operator_t>opABC,
            getComputeDesc(descCompute))
    check_status(status)
    return <intptr_t>desc

cpdef elementwiseTrinaryExecute(
        intptr_t handle, intptr_t plan,
        intptr_t alpha, intptr_t A,
        intptr_t beta, intptr_t B,
        intptr_t gamma, intptr_t C,
        intptr_t D):
    cdef intptr_t stream = stream_module.get_current_stream_ptr()
    with nogil:
        status = cutensorElementwiseTrinaryExecute(
            <Handle_t>handle, <Plan_t>plan,
            <void*>alpha, <void*>A,
            <void*>beta, <void*>B,
            <void*>gamma, <void*>C, <void*>D,
            <runtime.Stream>stream)
    check_status(status)

###############################################################################
# cutensorElementwiseBinary
###############################################################################

cpdef intptr_t createElementwiseBinary(
        intptr_t handle,
        intptr_t descA, intptr_t modeA, int opA,
        intptr_t descC, intptr_t modeC, int opC,
        intptr_t descD, intptr_t modeD,
        int opAB, int descCompute) except? 0:
    cdef OperationDescriptor_t desc
    with nogil:
        status = cutensorCreateElementwiseBinary(
            <Handle_t>handle, &desc,
            <TensorDescriptor_t>descA, <int32_t*>modeA, <Operator_t>opA,
            <TensorDescriptor_t>descC, <int32_t*>modeC, <Operator_t>opC,
            <TensorDescriptor_t>descD, <int32_t*>modeD,
            <Operator_t>opAB, getComputeDesc(descCompute))
    check_status(status)
    return <intptr_t>desc

cpdef elementwiseBinaryExecute(
        intptr_t handle, intptr_t plan,
        intptr_t alpha, intptr_t A,
        intptr_t gamma, intptr_t C, intptr_t D):
    cdef intptr_t stream = stream_module.get_current_stream_ptr()
    with nogil:
        status = cutensorElementwiseBinaryExecute(
            <Handle_t>handle, <Plan_t>plan,
            <void*>alpha, <void*>A,
            <void*>gamma, <void*>C, <void*>D,
            <runtime.Stream>stream)
    check_status(status)

###############################################################################
# cutensorPermute
###############################################################################

cpdef intptr_t createPermutation(
        intptr_t handle,
        intptr_t descA, intptr_t modeA, int opA,
        intptr_t descB, intptr_t modeB,
        int descCompute) except? 0:
    cdef OperationDescriptor_t desc
    with nogil:
        status = cutensorCreatePermutation(
            <Handle_t>handle, &desc,
            <TensorDescriptor_t>descA, <int32_t*>modeA, <Operator_t>opA,
            <TensorDescriptor_t>descB, <int32_t*>modeB,
            getComputeDesc(descCompute))
    check_status(status)
    return <intptr_t>desc

cpdef permute(
        intptr_t handle, intptr_t plan,
        intptr_t alpha, intptr_t A, intptr_t B):
    cdef intptr_t stream = stream_module.get_current_stream_ptr()
    with nogil:
        status =cutensorPermute(
            <Handle_t>handle, <Plan_t>plan,
            <void*>alpha, <void*>A,
            <void*>B, <runtime.Stream>stream)
    check_status(status)

###############################################################################
# cutensorContraction
###############################################################################

cpdef intptr_t createContraction(
        intptr_t handle,
        intptr_t descA, intptr_t modeA, int opA,
        intptr_t descB, intptr_t modeB, int opB,
        intptr_t descC, intptr_t modeC, int opC,
        intptr_t descD, intptr_t modeD,
        int descCompute) except? 0:
    cdef OperationDescriptor_t desc
    with nogil:
        status = cutensorCreateContraction(
            <Handle_t>handle, &desc,
            <TensorDescriptor_t>descA, <int32_t*>modeA, <Operator_t>opA,
            <TensorDescriptor_t>descB, <int32_t*>modeB, <Operator_t>opB,
            <TensorDescriptor_t>descC, <int32_t*>modeC, <Operator_t>opC,
            <TensorDescriptor_t>descD, <int32_t*>modeD,
            getComputeDesc(descCompute))
    check_status(status)
    return <intptr_t>desc

cpdef contract(
        intptr_t handle, intptr_t plan,
        intptr_t alpha, intptr_t A, intptr_t B,
        intptr_t beta, intptr_t C, intptr_t D,
        intptr_t workspace, uint64_t workspaceSize):
    cdef intptr_t stream = stream_module.get_current_stream_ptr()
    with nogil:
        status =cutensorContract(
            <Handle_t>handle, <Plan_t>plan,
            <void*>alpha, <void*>A, <void*>B,
            <void*>beta, <void*>C, <void*>D,
            <void*>workspace, workspaceSize, <runtime.Stream>stream)
    check_status(status)

###############################################################################
# cutensorReduction
###############################################################################

cpdef intptr_t createReduction(
        intptr_t handle,
        intptr_t descA, intptr_t modeA, int opA,
        intptr_t descC, intptr_t modeC, int opC,
        intptr_t descD, intptr_t modeD,
        int opReduce, int descCompute) except? 0:
    cdef OperationDescriptor_t desc
    with nogil:
        status = cutensorCreateReduction(
            <Handle_t>handle, &desc,
            <TensorDescriptor_t>descA, <int32_t*>modeA, <Operator_t>opA,
            <TensorDescriptor_t>descC, <int32_t*>modeC, <Operator_t>opC,
            <TensorDescriptor_t>descD, <int32_t*>modeD,
            <Operator_t>opReduce, getComputeDesc(descCompute))
    check_status(status)
    return <intptr_t>desc

cpdef reduce(
        intptr_t handle, intptr_t plan,
        intptr_t alpha, intptr_t A,
        intptr_t beta, intptr_t C, intptr_t D,
        intptr_t workspace, uint64_t workspaceSize):
    cdef intptr_t stream = stream_module.get_current_stream_ptr()
    with nogil:
        status = cutensorReduce(
            <Handle_t>handle, <Plan_t>plan,
            <void*>alpha, <void*>A,
            <void*>beta, <void*>C, <void*>D,
            <void*>workspace, workspaceSize, <runtime.Stream>stream)
    check_status(status)

#
cpdef destroyOperationDescriptor(intptr_t desc):
    with nogil:
        status = cutensorDestroyOperationDescriptor(
            <OperationDescriptor_t>desc)
    check_status(status)



from libc.stdint cimport intptr_t

cdef intptr_t get_current_stream_ptr()
cdef set_current_stream_ptr(intptr_t ptr, int device_id=*)
cpdef intptr_t get_default_stream_ptr()
cdef bint is_ptds_enabled()



# distutils: language = c++

"""Thin wrapper of cuRAND."""
cimport cython  # NOQA

from cupy_backends.cuda.api cimport runtime
from cupy_backends.cuda cimport stream as stream_module

###############################################################################
# Extern
###############################################################################

cdef extern from '../../cupy_rand.h' nogil:
    ctypedef void* Stream 'cudaStream_t'

    # Generator
    int curandCreateGenerator(Generator* generator, int rng_type)
    int curandDestroyGenerator(Generator generator)
    int curandGetVersion(int* version)

    # Stream
    int curandSetStream(Generator generator, Stream stream)
    int curandSetPseudoRandomGeneratorSeed(
        Generator generator, unsigned long long seed)
    int curandSetGeneratorOffset(
        Generator generator, unsigned long long offset)
    int curandSetGeneratorOrdering(Generator generator, Ordering order)

    # Generation functions
    int curandGenerate(
        Generator generator, unsigned int* outputPtr, size_t num)
    int curandGenerateLongLong(
        Generator generator, unsigned long long* outputPtr, size_t num)
    int curandGenerateUniform(
        Generator generator, float* outputPtr, size_t num)
    int curandGenerateUniformDouble(
        Generator generator, double* outputPtr, size_t num)
    int curandGenerateNormal(
        Generator generator, float* outputPtr, size_t num,
        float mean, float stddev)
    int curandGenerateNormalDouble(
        Generator generator, double* outputPtr, size_t n,
        double mean, double stddev)
    int curandGenerateLogNormal(
        Generator generator, float* outputPtr, size_t n,
        float mean, float stddev)
    int curandGenerateLogNormalDouble(
        Generator generator, double* outputPtr, size_t n,
        double mean, double stddev)
    int curandGeneratePoisson(
        Generator generator, unsigned int* outputPtr, size_t n, double lam)


###############################################################################
# Error handling
###############################################################################

STATUS = {
    0: 'CURAND_STATUS_SUCCESS',
    100: 'CURAND_STATUS_VERSION_MISMATCH',
    101: 'CURAND_STATUS_NOT_INITIALIZED',
    102: 'CURAND_STATUS_ALLOCATION_FAILED',
    103: 'CURAND_STATUS_TYPE_ERROR',
    104: 'CURAND_STATUS_OUT_OF_RANGE',
    105: 'CURAND_STATUS_LENGTH_NOT_MULTIPLE',
    106: 'CURAND_STATUS_DOUBLE_PRECISION_REQUIRED',
    201: 'CURAND_STATUS_LAUNCH_FAILURE',
    202: 'CURAND_STATUS_PREEXISTING_FAILURE',
    203: 'CURAND_STATUS_INITIALIZATION_FAILED',
    204: 'CURAND_STATUS_ARCH_MISMATCH',
    999: 'CURAND_STATUS_INTERNAL_ERROR',
}


class CURANDError(RuntimeError):

    def __init__(self, status):
        self.status = status
        super(CURANDError, self).__init__(STATUS[status])

    def __reduce__(self):
        return (type(self), (self.status,))


@cython.profile(False)
cpdef inline check_status(int status):
    if status != 0:
        raise CURANDError(status)


###############################################################################
# Generator
###############################################################################

cpdef size_t createGenerator(int rng_type) except? 0:
    cdef Generator generator
    with nogil:
        status = curandCreateGenerator(&generator, <RngType>rng_type)
    check_status(status)
    return <size_t>generator


cpdef destroyGenerator(size_t generator):
    status = curandDestroyGenerator(<Generator>generator)
    check_status(status)


cpdef int getVersion() except? -1:
    cdef int version
    status = curandGetVersion(&version)
    check_status(status)
    return version


cpdef setStream(size_t generator, size_t stream):
    # TODO(leofang): The support of stream capture is not mentioned at all in
    # the cuRAND docs (as of CUDA 11.5), so we disable this functionality.
    if not runtime._is_hip_environment and runtime.streamIsCapturing(stream):
        raise NotImplementedError(
            'calling cuRAND API during stream capture is currently '
            'unsupported')

    status = curandSetStream(<Generator>generator, <Stream>stream)
    check_status(status)


cdef _setStream(size_t generator):
    """Set current stream"""
    setStream(generator, stream_module.get_current_stream_ptr())


cpdef setPseudoRandomGeneratorSeed(size_t generator, unsigned long long seed):
    status = curandSetPseudoRandomGeneratorSeed(<Generator>generator, seed)
    check_status(status)


cpdef setGeneratorOffset(size_t generator, unsigned long long offset):
    status = curandSetGeneratorOffset(<Generator>generator, offset)
    check_status(status)


cpdef setGeneratorOrdering(size_t generator, int order):
    status = curandSetGeneratorOrdering(<Generator>generator, <Ordering>order)
    check_status(status)


###############################################################################
# Generation functions
###############################################################################

cpdef generate(size_t generator, size_t outputPtr, size_t num):
    _setStream(generator)
    status = curandGenerate(
        <Generator>generator, <unsigned int*>outputPtr, num)
    check_status(status)


cpdef generateLongLong(size_t generator, size_t outputPtr, size_t num):
    _setStream(generator)
    status = curandGenerateLongLong(
        <Generator>generator, <unsigned long long*>outputPtr, num)
    check_status(status)


cpdef generateUniform(size_t generator, size_t outputPtr, size_t num):
    _setStream(generator)
    status = curandGenerateUniform(
        <Generator>generator, <float*>outputPtr, num)
    check_status(status)


cpdef generateUniformDouble(size_t generator, size_t outputPtr, size_t num):
    _setStream(generator)
    status = curandGenerateUniformDouble(
        <Generator>generator, <double*>outputPtr, num)
    check_status(status)


cpdef generateNormal(size_t generator, size_t outputPtr, size_t n,
                     float mean, float stddev):
    if n % 2 == 1:
        msg = ('curandGenerateNormal can only generate even number of '
               'random variables simultaneously. See issue #390 for detail.')
        raise ValueError(msg)
    _setStream(generator)
    status = curandGenerateNormal(
        <Generator>generator, <float*>outputPtr, n, mean, stddev)
    check_status(status)


cpdef generateNormalDouble(size_t generator, size_t outputPtr, size_t n,
                           float mean, float stddev):
    if n % 2 == 1:
        msg = ('curandGenerateNormalDouble can only generate even number of '
               'random variables simultaneously. See issue #390 for detail.')
        raise ValueError(msg)
    _setStream(generator)
    status = curandGenerateNormalDouble(
        <Generator>generator, <double*>outputPtr, n, mean, stddev)
    check_status(status)


def generateLogNormal(size_t generator, size_t outputPtr, size_t n,
                      float mean, float stddev):
    if n % 2 == 1:
        msg = ('curandGenerateLogNormal can only generate even number of '
               'random variables simultaneously. See issue #390 for detail.')
        raise ValueError(msg)
    _setStream(generator)
    status = curandGenerateLogNormal(
        <Generator>generator, <float*>outputPtr, n, mean, stddev)
    check_status(status)


cpdef generateLogNormalDouble(size_t generator, size_t outputPtr, size_t n,
                              float mean, float stddev):
    if n % 2 == 1:
        msg = ('curandGenerateLogNormalDouble can only generate even number '
               'of random variables simultaneously. See issue #390 for '
               'detail.')
        raise ValueError(msg)
    _setStream(generator)
    status = curandGenerateLogNormalDouble(
        <Generator>generator, <double*>outputPtr, n, mean, stddev)
    check_status(status)


cpdef generatePoisson(size_t generator, size_t outputPtr, size_t n,
                      double lam):
    _setStream(generator)
    status = curandGeneratePoisson(
        <Generator>generator, <unsigned int*>outputPtr, n, lam)
    check_status(status)
